# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
from metku.frame2d.frame2d import *
from metku.optimization.structopt import *
from metku.sections.steel.catalogue import rhs_profiles, ipe_profiles
#from metku.truss2d import *
from metku.optimization.solvers import *

RHS_PROFILES = list(rhs_profiles.keys())

# Sorted by area
RHS_PROFILES = ['RHS 40X40X2.0', 'RHS 40X40X2.5', 'RHS 50X50X2.0',
                'RHS 40X40X3.0', 'RHS 60X60X2.0', 'RHS 50X50X2.5',
                'RHS 70X70X2.0', 'RHS 50X50X3.0', 'RHS 60X60X2.5',
                'RHS 80X80X2.0', 'RHS 70X70X2.5', 'RHS 60X60X3.0',
                'RHS 50X50X4.0', 'RHS 80X80X2.5', 'RHS 70X70X3.0',
                'RHS 60X60X4.0', 'RHS 90X90X2.5', 'RHS 80X80X3.0',
                'RHS 100X100X2.5', 'RHS 70X70X4.0', 'RHS 90X90X3.0',
                'RHS 60X60X5.0', 'RHS 100X100X3.0', 'RHS 120X120X2.5',
                'RHS 80X80X4.0', 'RHS 70X70X5.0', 'RHS 90X90X4.0',
                'RHS 140X140X2.5', 'RHS 120X120X3.0', 'RHS 80X80X5.0',
                'RHS 150X150X2.5', 'RHS 100X100X4.0', 'RHS 140X140X3.0',
                'RHS 90X90X5.0', 'RHS 150X150X3.0', 'RHS 120X120X4.0',
                'RHS 100X100X5.0', 'RHS 160X160X3.0', 'RHS 140X140X4.0',
                'RHS 100X100X6.0', 'RHS 120X120X5.0', 'RHS 150X150X4.0',
                'RHS 160X160X4.0', 'RHS 140X140X5.0', 'RHS 120X120X6.0',
                'RHS 180X180X4.0', 'RHS 150X150X5.0', 'RHS 120X120X7.1',
                'RHS 160X160X5.0', 'RHS 200X200X4.0', 'RHS 140X140X6.0',
                'RHS 150X150X6.0', 'RHS 120X120X8.0', 'RHS 180X180X5.0',
                'RHS 160X160X6.0', 'RHS 120X120X8.8', 'RHS 180X180X5.6',
                'RHS 200X200X5.0', 'RHS 150X150X7.1', 'RHS 120X120X10.0',
                'RHS 180X180X6.0', 'RHS 160X160X7.1', 'RHS 220X220X5.0',
                'RHS 150X150X8.0', 'RHS 200X200X6.0', 'RHS 160X160X8.0',
                'RHS 150X150X8.8', 'RHS 180X180X7.1', 'RHS 250X250X5.0',
                'RHS 220X220X6.0', 'RHS 160X160X8.8', 'RHS 150X150X10.0',
                'RHS 180X180X8.0', 'RHS 200X200X7.1', 'RHS 160X160X10.0',
                'RHS 180X180X8.8', 'RHS 250X250X6.0', 'RHS 300X300X5.0',
                'RHS 220X220X7.1', 'RHS 200X200X8.0', 'RHS 260X260X6.0',
                'RHS 180X180X10.0', 'RHS 200X200X8.8', 'RHS 220X220X8.0',
                'RHS 250X250X7.1', 'RHS 300X300X6.0', 'RHS 260X260X7.1',
                'RHS 220X220X8.8', 'RHS 200X200X10.0', 'RHS 250X250X8.0',
                'RHS 180X180X12.5', 'RHS 260X260X8.0', 'RHS 220X220X10.0',
                'RHS 300X300X7.1', 'RHS 250X250X8.8', 'RHS 260X260X8.8',
                'RHS 200X200X12.5', 'RHS 300X300X8.0', 'RHS 250X250X10.0',
                'RHS 400X400X6.0', 'RHS 260X260X10.0', 'RHS 300X300X8.8',
                'RHS 400X400X7.1', 'RHS 250X250X12.5', 'RHS 300X300X10.0',
                'RHS 260X260X12.5', 'RHS 400X400X8.0', 'RHS 400X400X8.8',
                'RHS 300X300X12.5', 'RHS 400X400X10.0', 'RHS 400X400X12.5']

RHS_A = [293.6991118430775, 358.9048622548086, 373.6991118430775,
         420.8230016469244, 453.6991118430775, 458.9048622548086,
         533.6991118430775, 540.8230016469245, 558.9048622548087,
         613.6991118430775, 658.9048622548087, 660.8230016469245,
         694.79644737231, 758.9048622548087, 780.8230016469245,
         854.79644737231, 858.9048622548087, 900.8230016469245,
         958.9048622548087, 1014.79644737231, 1020.8230016469245,
         1035.6194490192345, 1140.8230016469245, 1158.9048622548087,
         1174.79644737231, 1235.6194490192345, 1334.79644737231,
         1358.9048622548087, 1380.8230016469245, 1435.6194490192345,
         1458.9048622548087, 1494.79644737231, 1620.8230016469245,
         1635.6194490192345, 1740.8230016469245, 1814.79644737231,
         1835.6194490192345, 1860.8230016469245, 2134.79644737231,
         2163.292006587698, 2235.6194490192347, 2294.79644737231,
         2454.79644737231, 2635.6194490192347, 2643.292006587698,
         2774.79644737231, 2835.6194490192347, 3033.2707426698457,
         3035.6194490192347, 3094.79644737231, 3123.292006587698,
         3363.292006587698, 3364.247719318987, 3435.6194490192347,
         3603.292006587698, 3648.3397403759745, 3825.8010368497276,
         3835.6194490192347, 3885.2707426698457, 4056.6370614359175,
         4083.292006587698, 4169.270742669845, 4235.619449019235,
         4324.247719318987, 4563.292006587698, 4644.247719318987,
         4704.3397403759745, 4737.270742669846, 4835.619449019235,
         5043.292006587698, 5056.3397403759745, 5256.6370614359175,
         5284.247719318987, 5305.270742669846, 5656.6370614359175,
         5760.339740375975, 5763.292006587698, 5835.619449019235,
         5873.270742669846, 5924.247719318987, 6003.292006587698,
         6456.6370614359175, 6464.339740375975, 6564.247719318987,
         6725.270742669845, 6963.292006587698, 7009.270742669845,
         7168.339740375975, 7256.6370614359175, 7524.247719318987,
         7704.369260617026, 7844.247719318987, 8056.6370614359175,
         8145.270742669845, 8224.339740375975, 8576.339740375975,
         8704.369260617026, 9124.247719318988, 9256.637061435917,
         9363.292006587697, 9656.637061435917, 9984.339740375975,
         10985.270742669845, 11204.369260617026, 11256.637061435917,
         11704.369260617026, 12324.247719318988, 13504.339740375975,
         13704.369260617026, 15256.637061435917, 18704.369260617026]

RHS_Iy = [69402.1348577099, 82151.28824369762, 141469.80388201258,
          93235.56709132508, 251422.42849846912, 169438.8058049558,
          407260.0087070795, 194671.3623630676, 303421.5664789544,
          616982.5445078439, 494099.5702656935, 351348.30771715636,
          237358.76890471048, 751472.8171651728, 575266.4031535913,
          435510.7428089776, 1085541.3071773928, 878425.6486723726,
          1506305.0403023532, 721202.5390818602, 1272826.0442734999,
          504944.22072287824, 1770467.5899569737, 2647918.2358904947,
          1110434.1577233584, 846291.9300855394, 1619205.598733472,
          4256312.403929599, 3123474.1315709595, 1314420.611899162,
          5260552.35261826, 2263516.8621122013, 5033445.27351433,
          1929330.2661637464, 6227292.569609535, 4022758.855975506,
          2711020.892879293, 7596381.015787086, 6516160.139313272,
          3114741.7978090816, 4854745.06366327, 8078170.514535079,
          9871720.712125503, 7905593.124251096, 5621572.923474502,
          14217440.574412191, 9821188.61322145, 6235230.294796982,
          12023565.074642764, 19681319.726173345, 9204262.450457461,
          11459054.114443017, 6768760.707404428, 17368660.914838284,
          14054810.378757961, 7198754.151445843, 19184945.76423401,
          24100880.64483765, 12896997.263231725, 7768082.442480799,
          20365216.708375998, 15874082.660310294, 32380224.26464086,
          14118333.707433986, 28327481.43931158, 17412349.479375735,
          15131229.417059045, 23133398.0658679, 48050096.98772789,
          38133604.5715647, 18695912.47963438, 16525294.695811596,
          25458618.181157082, 32322239.619959485, 20476695.81973211,
          27417265.565841436, 56720023.77241476, 84168837.64189216,
          43667807.32258503, 35662536.426802225, 64045426.04002355,
          30167993.626788545, 38495934.600123696, 48280104.21631118,
          65227020.405024536, 99636681.11375256, 73737921.17343803,
          52213519.582481146, 42510618.84613217, 72292048.7953192,
          34064339.66473276, 81780188.42692047, 57824571.47776296,
          115161339.61842687, 78353864.45865832, 88691837.39142165,
          48594155.981109016, 128006870.81298491, 87066739.32324764,
          241042340.82113203, 98646458.97788613, 139105018.9926629,
          280318593.3302435, 101613144.87508953, 155193656.12715802,
          115478848.04297818, 312692443.7957626, 340887002.087082,
          183481345.34484133, 382159878.71536344, 458765381.0116587]

IPE_PROFILES = list(ipe_profiles.keys())
ten_bar_DLM = [41, 0, 38, 31, 0, 0, 27, 38, 37, 0]
ten_bar_PSO = [40, 0, 38, 29, 0, 0, 27, 39, 37, 1]

TEN_BAR_AREAS_in2 = [1.62, 1.80, 1.99, 2.13, 2.38, 2.62, 2.63, 2.88, 2.93,
                     3.09, 3.13, 3.38, 3.47, 3.55, 3.63, 3.84, 3.87, 3.88,
                     4.18, 4.22, 4.49, 4.59, 4.80, 4.97, 5.12, 5.74, 7.22,
                     7.97, 11.50, 13.50, 13.90, 14.20, 15.50, 16.00, 16.90,
                     18.80, 19.90, 22.00, 22.90, 26.50, 30.00, 33.50, ]

TEN_BAR_AREAS_mm2 = [1044.9, 1161.0, 1283.55, 1373.85, 1535.1, 1689.9,
                     1696.35, 1857.6, 1889.85, 1993.05, 2018.85, 2180.1,
                     2238.15, 2289.75, 2341.35, 2476.8, 2496.15, 2502.6,
                     2696.1, 2721.9, 2896.05, 2960.55, 3096.0, 3205.65,
                     3302.4, 3702.3, 4656.9, 5140.65, 7417.5, 8707.5, 8965.5,
                     9159.0, 9997.5, 10320.0, 10900.5, 12126.0, 12835.5,
                     14190.0, 14770.5, 17092.5, 19350.0, 21607.5]

FIFTEEN_BAR_AREAS_in2 = [0.111, 0.141, 0.174, 0.220, 0.270, 0.287, 0.347,
                         0.440,
                         0.539, 0.954, 1.081, 1.174, 1.333, 1.488, 1.764,
                         2.142,
                         2.697, 2.800, 3.131, 3.565, 3.813, 4.805, 5.952,
                         6.572,
                         7.192, 8.525, 9.300, 10.850, 13.300, 14.290, 17.170,
                         19.180]

in2_to_mm2 = 645.16

FIFTEEN_BAR_AREAS_mm2 = [71.613, 90.968, 112.258, 141.935, 174.193, 185.161,
                         223.871,
                         283.87, 347.741, 615.483, 697.418, 757.418, 859.998,
                         959.998,
                         1138.062, 1381.933, 1739.997, 1806.448, 2019.996,
                         2299.995,
                         2459.995, 3099.994, 3839.992, 4239.992, 4639.991,
                         5499.989,
                         5999.988, 6999.986, 8580.628, 9219.336, 11077.397,
                         12374.169]

THREE_BAR_AREAS_mm2 = FIFTEEN_BAR_AREAS_mm2


class DiscreteFrame(OptimizationProblem):
    """
    General class for creating optimization problems that utilize the Frame2D -module
    
    Finite element analysis is done in the objective function and it's should always
    be calculated before constraints.

    """

    def __init__(self, name="Discrete Frame", profiles=RHS_PROFILES):
        super().__init__(name)

        self.profiles = profiles

    def add_structure(self, structure):
        self.structure = structure
        self.create_variables()
        self.create_objective()

    def create_variables(self):
        self.vars = []
        for i in range(len(self.structure.members)):
            X = IntegerVariable('X' + str(i + 1), 0, len(self.profiles) - 1)
            self.vars.append(X)

    def create_objective(self):

        def objective(X):
            if isinstance(X[0], IntegerVariable):
                X = [x.value for x in X]

            weight = 0
            for x, mem in zip(X, self.structure.members.values()):
                x = int(x)
                mem.profile = self.profiles[x]
                weight += mem.weight
            # Calculate stucture with new values
            self.structure.calculate()
            return weight

        self.obj = objective

    def _create_normal_force_constraints(self):
        """
        Creates normal force constraints
        """
        for i in range(len(self.structure.members)):
            def con_fun(X, i=i):
                mem = self.structure.members[i]
                return mem.ned / mem.NRd - 1

            constr = NonLinearConstraint(con_fun,
                                         name='Normal Force ' + str(i + 1))
            self.cons.append(constr)

    def _create_shear_force_constraints(self):
        """
        Creates shear force constraints
        """
        for i in range(len(self.structure.members)):
            def con_fun(X, i=i):
                mem = self.structure.members[i]
                return mem.ved / mem.VRd - 1

            constr = NonLinearConstraint(con_fun,
                                         name='Shear Force ' + str(i + 1))
            self.cons.append(constr)

    def _create_bending_moment_constraints(self):
        """
        Creates bending moment constraints about y-axis
        """
        for i in range(len(self.structure.members)):
            def con_fun(X, i=i):
                mem = self.structure.members[i]
                return mem.med / mem.MRd[0] - 1

            constr = NonLinearConstraint(con_fun,
                                         name='Bending Moment ' + str(i + 1))
            self.cons.append(constr)

    def create_constraints(self, N=True, V=True, M=True):

        # Initialize constraints as an empty list
        self.cons = []
        # Create normal force constraints
        if N:
            self._create_normal_force_constraints()
        if V:
            self._create_shear_force_constraints()
        if M:
            self._create_bending_moment_constraints()


class ThreeBarTruss(OptimizationProblem):

    # Problem parameters
    L = 3048  # mm
    F = 444890  # N # 1e5 lb
    E = 68950  # MPa # 10e7 psi
    rho = 2768e-9  # kg/mm3

    properties = {
        'L': str(L) + " mm",
        'F': f'{F / 1e3 :.2f} kN',
        'E': f'{E / 1e3 :.2f} MPa',
        'rho': f'{rho * 1e9 :.0f} kg/m3'
    }

    # Constraint limits
    sigma_max = 172.37  # MPa # 25000 psi
    delta_max = 50.8  # mm

    def __init__(self, prob_type="discrete"):

        super().__init__(name='ThreeBarTruss')

        self.prob_type = prob_type.lower()
        self.structure = self.create_structure()
        self.create_constraints()
        self.create_objective()

    def create_objective(self):

        def objective(X):
            if np.any(self.X != X):
                self.substitute_variables(X)
            weight = 0
            for x, mem in zip(X, self.structure.members.values()):
                if self.prob_type == 'discrete':
                    A = THREE_BAR_AREAS_mm2[x]
                else:
                    A = x
                weight += self.rho * A * mem.length
            return weight

        self.obj = objective

    def create_structure(self):

        frame = Frame2D(num_elements=1)
        # Nodes
        n1 = [0, 0]
        n2 = [self.L, 0]
        n3 = [2 * self.L, 0]
        n4 = [self.L, -self.L]
        # Create bars
        bar1 = SteelBeam([n1, n4])
        bar2 = SteelBeam([n2, n4])
        bar3 = SteelBeam([n3, n4])
        # Add bars
        frame.add(bar1)
        frame.add(bar2)
        frame.add(bar3)
        # Supports
        frame.add(XYHingedSupport(n1))
        frame.add(XYHingedSupport(n2))
        frame.add(XYHingedSupport(n3))
        # Point loads
        frame.add(PointLoad(n4, [-self.F, -self.F, 0]))
        # Change material properties
        for mem in frame.members.values():
            mem.fy = self.sigma_max
            mem.E = self.E  # MPa
            mem.A = THREE_BAR_AREAS_mm2[0]
            mem.Sj1 = 0
            mem.Sj2 = 0
        frame.generate()
        frame.calculate()

        if self.prob_type == "discrete":
            # Variables
            A1 = DiscreteVariable('A1', profiles=THREE_BAR_AREAS_mm2,
                                  target={"property": "A",
                                          "objects": [bar1, bar3]})
            A2 = DiscreteVariable('A2', profiles=THREE_BAR_AREAS_mm2,
                                  target={"property": "A", "objects": [bar2]})
            vars = [A1, A2]

        elif self.prob_type == "continuous":
            A1 = Variable('A1', lb=THREE_BAR_AREAS_mm2[0],
                          ub=THREE_BAR_AREAS_mm2[-1],
                          target={"property": "A", "objects": [bar1, bar3]})
            A2 = Variable('A2', lb=THREE_BAR_AREAS_mm2[0],
                          ub=THREE_BAR_AREAS_mm2[-1],
                          target={"property": "A", "objects": [bar2]})

            vars = [A1, A2]

        self.vars = vars

        return frame


    def _create_stress_constraint(self, mem, i, j):
        """
        THIS DOES NOT WORK YET!
        GIVES DIFFERENT SOLUTIONS THAN THE FUNCTION CREATED
        IN THE LOOP!
        :param mem:
        :param i:
        :param j:
        :return:
        """
        def con_fun(x, i=i, j=j):

            if self.prob_type == 'discrete':
                A = TEN_BAR_AREAS_mm2[x[j]]
            else:
                A = x[j]

            return mem.ned / (A * mem.fy) - 1

        return con_fun

    def create_constraints(self):

        # Initialize constraints as an empty list
        self.cons = []

        i = 0
        for j, var in enumerate(self.vars):
            for mem in var.target["objects"]:
                if isinstance(mem, FrameMember):
                    i += 1
                    def stress_fun(x, i=i, j=j):

                        if self.prob_type == 'discrete':
                            A = TEN_BAR_AREAS_mm2[x[j]]
                        else:
                            A = x[j]

                        return mem.ned / (A * mem.fy) - 1

                    def buckling_fun(x, i=i, j=j):

                        if self.prob_type == 'discrete':
                            A = TEN_BAR_AREAS_mm2[x[j]]
                        else:
                            A = x[j]

                        sigma_cr = 100 * mem.E * A / (8 * mem.length ** 2)
                        sigma = -mem.ned / A
                        return sigma / sigma_cr - 1

                    def disp_fun(A, i=i):
                        displacements = mem.nodal_displacements.values()
                        max_vals = [max(l[0:2]) for l in displacements]
                        min_vals = [min(l[0:2]) for l in displacements]
                        max_val = max(max_vals)
                        min_val = min(min_vals)
                        abs_max = max(max_val, abs(min_val))
                        return abs_max / self.delta_max - 1

                    stress_con = NonLinearConstraint(con_fun=stress_fun,
                                                 name="Stress " + str(i),
                                                 problem=self)
                    stress_con.fea_required = True

                    buckling_con = NonLinearConstraint(con_fun=buckling_fun,
                                                     name="Buckling " + str(i),
                                                     problem=self)
                    buckling_con.fea_required = True

                    disp_con = NonLinearConstraint(con_fun=disp_fun,
                                                 name='Displacement ' + str(i),
                                                 problem=self)
                    disp_con.fea_required = True

                    self.cons.append(stress_con)
                    self.cons.append(buckling_con)
                    self.cons.append(disp_con)












        # # Create stress constraints
        # for i in range(len(self.structure.members)):
        #     def con_fun(x, i=i):
        #         mem = self.structure.members[i]
        #         mem.A = x[i]
        #         return abs(mem.ned / mem.NRd) - 1
        #
        #     constr = NonLinearConstraint(con_fun, name='Stress ' + str(i + 1),
        #                                  parent=self)
        #     constr.fea_required = True
        #     self.cons.append(constr)
        #
        # # Create displacement constraints
        # for j in range(len(self.structure.members)):
        #     def con_fun(A, j=j):
        #         mem = self.structure.members[j]
        #         displacements = mem.nodal_displacements.values()
        #         max_vals = [max(l[0:2]) for l in displacements]
        #         min_vals = [min(l[0:2]) for l in displacements]
        #         max_val = max(max_vals)
        #         min_val = min(min_vals)
        #         abs_max = max(max_val, abs(min_val))
        #         return abs_max / self.delta_max - 1
        #
        #     constr = NonLinearConstraint(con_fun,
        #                                  name='Displacement ' + str(1 + j),
        #                                  parent=self)
        #     constr.fea_required = True
        #     self.cons.append(constr)
        #
        # # Create buckling constraints
        # for j in range(len(self.structure.members)):
        #     def con_fun(x, j=j):
        #         mem = self.structure.members[j]
        #         sigma_cr = 100 * mem.E * x[j] / (8 * mem.length ** 2)
        #         sigma = abs(mem.ned) / mem.A
        #         return sigma / sigma_cr - 1
        #
        #     constr = NonLinearConstraint(con_fun,
        #                                  name='Buckling ' + str(1 + j),
        #                                  parent=self)
        #     constr.fea_required = True
        #     self.cons.append(constr)


class FifteenBarTruss(OptimizationProblem):
    # Problem parameters
    L = 3048  # mm
    F = 444890  # N # 1e5 lb
    E = 68950  # MPa # 10e7 psi
    rho = 2768e-9  # kg/mm3

    properties = {
        'L': str(L) + " mm",
        'F': f'{F / 1e3 :.2f} kN',
        'E': f'{E / 1e3 :.2f} MPa',
        'rho': f'{rho * 1e9 :.0f} kg/m3'
    }

    # Variables
    A1 = IntegerVariable('A1', 0, len(FIFTEEN_BAR_AREAS_mm2) - 1)
    A2 = IntegerVariable('A2', 0, len(FIFTEEN_BAR_AREAS_mm2) - 1)
    A3 = IntegerVariable('A3', 0, len(FIFTEEN_BAR_AREAS_mm2) - 1)
    A4 = IntegerVariable('A4', 0, len(FIFTEEN_BAR_AREAS_mm2) - 1)
    A5 = IntegerVariable('A5', 0, len(FIFTEEN_BAR_AREAS_mm2) - 1)
    A6 = IntegerVariable('A6', 0, len(FIFTEEN_BAR_AREAS_mm2) - 1)
    A7 = IntegerVariable('A7', 0, len(FIFTEEN_BAR_AREAS_mm2) - 1)
    A8 = IntegerVariable('A8', 0, len(FIFTEEN_BAR_AREAS_mm2) - 1)
    A9 = IntegerVariable('A9', 0, len(FIFTEEN_BAR_AREAS_mm2) - 1)
    A10 = IntegerVariable('A10', 0, len(FIFTEEN_BAR_AREAS_mm2) - 1)
    A11 = IntegerVariable('A11', 0, len(FIFTEEN_BAR_AREAS_mm2) - 1)
    A12 = IntegerVariable('A12', 0, len(FIFTEEN_BAR_AREAS_mm2) - 1)
    A13 = IntegerVariable('A13', 0, len(FIFTEEN_BAR_AREAS_mm2) - 1)
    A14 = IntegerVariable('A14', 0, len(FIFTEEN_BAR_AREAS_mm2) - 1)
    A15 = IntegerVariable('A15', 0, len(FIFTEEN_BAR_AREAS_mm2) - 1)

    int_vars = [A1, A2, A3, A4, A5,
                A6, A7, A8, A9, A10,
                A11, A12, A13, A14, A15]

    # Constraint limits

    sigma_max = 172.37  # MPa # 25000 psi
    delta_max = 50.8  # mm

    def __init__(self):

        super().__init__(name='TenBarTruss',
                         structure=self.create_structure(),
                         variables=self.int_vars)
        self.create_constraints()
        self.create_objective()

    def create_objective(self):

        def objective(X):
            if isinstance(X[0], IntegerVariable):
                X = [x.value for x in X]

            weight = 0
            for x, mem in zip(X, self.structure.members.values()):
                x = int(x)
                mem.A = TEN_BAR_AREAS_mm2[x]
                weight += self.rho * mem.A * mem.length
            # Calculate stucture with new values
            self.structure.calculate()
            return weight

        self.obj = objective

    def create_structure(self):

        frame = Frame2D(num_elements=1)
        # Nodes
        n1 = [0, self.L]
        n2 = [self.L, self.L]
        n3 = [2 * self.L, self.L]
        n4 = [3 * self.L, self.L]
        n5 = [0, 0]
        n6 = [self.L, 0]
        n7 = [2 * self.L, 0]
        n8 = [3 * self.L, 0]
        # Create bars
        bar1 = SteelBeam([n1, n2])
        bar2 = SteelBeam([n2, n3])
        bar3 = SteelBeam([n3, n4])
        bar4 = SteelBeam([n5, n6])
        bar5 = SteelBeam([n6, n7])
        bar6 = SteelBeam([n7, n8])
        bar7 = SteelBeam([n6, n2])
        bar8 = SteelBeam([n7, n3])
        bar9 = SteelBeam([n8, n4])
        bar10 = SteelBeam([n1, n6])
        bar11 = SteelBeam([n5, n2])
        bar12 = SteelBeam([n2, n7])
        bar13 = SteelBeam([n6, n3])
        bar14 = SteelBeam([n3, n8])
        bar15 = SteelBeam([n7, n4])
        # Add bars
        frame.add(bar1)
        frame.add(bar2)
        frame.add(bar3)
        frame.add(bar4)
        frame.add(bar5)
        frame.add(bar6)
        frame.add(bar7)
        frame.add(bar8)
        frame.add(bar9)
        frame.add(bar10)
        frame.add(bar11)
        frame.add(bar12)
        frame.add(bar13)
        frame.add(bar14)
        frame.add(bar15)
        # Supports
        frame.add(XYHingedSupport(n1))
        frame.add(XYHingedSupport(n5))
        # Point loads
        frame.add(PointLoad(n8, [0, -self.F, 0]))
        # Change material properties
        for mem in frame.members.values():
            mem.cross_section.fy = self.sigma_max
            mem.E = self.E  # MPa
            mem.A = FIFTEEN_BAR_AREAS_mm2[0]
            mem.Sj1 = 0
            mem.Sj2 = 0
        frame.generate()
        frame.calculate()
        return frame

    def create_constraints(self):

        # Initialize constraints as an empty list
        self.cons = []
        # Create stress constraints
        for i in range(len(self.structure.members)):
            def con_fun(A, i=i):
                mem = self.structure.members[i]
                mem.A = TEN_BAR_AREAS_mm2[int(A[i])]
                return mem.ned / mem.NRd - 1

            constr = NonLinearConstraint(con_fun, name='Stress ' + str(i + 1))
            self.cons.append(constr)

        # Create buckling constraints
        for j in range(len(self.structure.members)):
            def con_fun(A, j=j):
                mem = self.structure.members[j]
                sigma_cr = 100 * mem.E * mem.A / (8 * mem.length ** 2)
                sigma = abs(mem.ned) / mem.A
                return sigma / sigma_cr - 1

            constr = NonLinearConstraint(con_fun,
                                         name='Buckling ' + str(11 + j))
            self.cons.append(constr)


class TenBarTruss(OptimizationProblem):
    # Problem parameters
    L = 9144  # mm
    F = 444890  # N
    E = 68950  # MPa
    rho = 2768e-9  # kg/mm3

    properties = {
        'L': str(L) + " mm",
        'F': f'{F / 1e3 :.2f} kN',
        'E': f'{E / 1e3 :.2f} MPa',
        'rho': f'{rho * 1e9 :.0f} kg/m3'
    }

    # Constraint limits
    sigma_max = 172.37  # MPa
    delta_max = 50.8  # mm

    def __init__(self, prob_type="discrete"):

        self.prob_type = prob_type
        super().__init__(name='TenBarTruss')
        self.structure = self.create_structure()
        self.create_constraints()
        self.create_objective()

    def create_structure(self):

        frame = Frame2D(num_elements=1)
        # Members
        n1 = [2 * self.L, self.L]
        n2 = [2 * self.L, 0]
        n3 = [self.L, self.L]
        n4 = [self.L, 0]
        n5 = [0, 0]
        n6 = [0, self.L]
        bar1 = SteelBeam([n6, n3])
        bar2 = SteelBeam([n3, n1])
        bar3 = SteelBeam([n5, n4])
        bar4 = SteelBeam([n4, n2])
        bar5 = SteelBeam([n3, n4])
        bar6 = SteelBeam([n1, n2])
        bar7 = SteelBeam([n6, n4])
        bar8 = SteelBeam([n5, n3])
        bar9 = SteelBeam([n3, n2])
        bar10 = SteelBeam([n4, n1])
        frame.add(bar1)
        frame.add(bar2)
        frame.add(bar3)
        frame.add(bar4)
        frame.add(bar5)
        frame.add(bar6)
        frame.add(bar7)
        frame.add(bar8)
        frame.add(bar9)
        frame.add(bar10)
        # Supports
        frame.add(XYHingedSupport(n5))
        frame.add(XYHingedSupport(n6))
        # Point loads
        frame.add(PointLoad(n4, [0, -self.F, 0]))
        frame.add(PointLoad(n2, [0, -self.F, 0]))
        # Change material properties
        for mem in frame.members.values():
            mem.cross_section.fy = self.sigma_max
            mem.E = self.E  # MPa
            mem.A = TEN_BAR_AREAS_mm2[0]
            mem.Sj1 = 0
            mem.Sj2 = 0
        frame.generate()
        frame.calculate()

        if self.prob_type == "discrete":
            # Variables
            A1 = DiscreteVariable('A1', profiles=TEN_BAR_AREAS_mm2,
                                  target={"property": "A", "objects": [bar1]})
            A2 = DiscreteVariable('A2', profiles=TEN_BAR_AREAS_mm2,
                                  target={"property": "A", "objects": [bar2]})
            A3 = DiscreteVariable('A3', profiles=TEN_BAR_AREAS_mm2,
                                  target={"property": "A", "objects": [bar3]})
            A4 = DiscreteVariable('A4', profiles=TEN_BAR_AREAS_mm2,
                                  target={"property": "A", "objects": [bar4]})
            A5 = DiscreteVariable('A5', profiles=TEN_BAR_AREAS_mm2,
                                  target={"property": "A", "objects": [bar5]})
            A6 = DiscreteVariable('A6', profiles=TEN_BAR_AREAS_mm2,
                                  target={"property": "A", "objects": [bar6]})
            A7 = DiscreteVariable('A7', profiles=TEN_BAR_AREAS_mm2,
                                  target={"property": "A", "objects": [bar7]})
            A8 = DiscreteVariable('A8', profiles=TEN_BAR_AREAS_mm2,
                                  target={"property": "A", "objects": [bar8]})
            A9 = DiscreteVariable('A9', profiles=TEN_BAR_AREAS_mm2,
                                  target={"property": "A", "objects": [bar9]})
            A10 = DiscreteVariable('A10', profiles=TEN_BAR_AREAS_mm2,
                                   target={"property": "A",
                                           "objects": [bar10]})

            vars = [A1, A2, A3, A4, A5, A6, A7, A8, A9, A10]


        elif self.prob_type == "continuous":
            A1 = Variable('A1', lb=TEN_BAR_AREAS_mm2[0],
                          ub=TEN_BAR_AREAS_mm2[-1],
                          target={"property": "A", "objects": [bar1]})
            A2 = Variable('A2', lb=TEN_BAR_AREAS_mm2[0],
                          ub=TEN_BAR_AREAS_mm2[-1],
                          target={"property": "A", "objects": [bar2]})
            A3 = Variable('A3', lb=TEN_BAR_AREAS_mm2[0],
                          ub=TEN_BAR_AREAS_mm2[-1],
                          target={"property": "A", "objects": [bar3]})
            A4 = Variable('A4', lb=TEN_BAR_AREAS_mm2[0],
                          ub=TEN_BAR_AREAS_mm2[-1],
                          target={"property": "A", "objects": [bar4]})
            A5 = Variable('A5', lb=TEN_BAR_AREAS_mm2[0],
                          ub=TEN_BAR_AREAS_mm2[-1],
                          target={"property": "A", "objects": [bar5]})
            A6 = Variable('A6', lb=TEN_BAR_AREAS_mm2[0],
                          ub=TEN_BAR_AREAS_mm2[-1],
                          target={"property": "A", "objects": [bar6]})
            A7 = Variable('A7', lb=TEN_BAR_AREAS_mm2[0],
                          ub=TEN_BAR_AREAS_mm2[-1],
                          target={"property": "A", "objects": [bar7]})
            A8 = Variable('A8', lb=TEN_BAR_AREAS_mm2[0],
                          ub=TEN_BAR_AREAS_mm2[-1],
                          target={"property": "A", "objects": [bar8]})
            A9 = Variable('A9', lb=TEN_BAR_AREAS_mm2[0],
                          ub=TEN_BAR_AREAS_mm2[-1],
                          target={"property": "A", "objects": [bar9]})
            A10 = Variable('A10', lb=TEN_BAR_AREAS_mm2[0],
                           ub=TEN_BAR_AREAS_mm2[-1],
                           target={"property": "A",
                                   "objects": [bar10]})

            vars = [A1, A2, A3, A4, A5, A6, A7, A8, A9, A10]

        else:
            raise TypeError("Problem type must be either 'dicrete' "
                            "or 'continuous")

        self.vars = vars

        return frame

    def create_constraints(self):
        """
        Creates constraints for the problem
        """
        # Initialize constraints as an empty list
        self.cons = []
        # Create stress constraints
        for i in range(len(self.structure.members)):
            def con_fun(A, i=i):
                mem = self.structure.members[i]
                return mem.ned / mem.NRd - 1

            constr = NonLinearConstraint(con_fun, name='Stress ' + str(i + 1),
                                         parent=self)
            constr.fea_required = True
            self.cons.append(constr)

        # Create displacement constraints
        for j in range(len(self.structure.members)):
            def con_fun(A, j=j):
                mem = self.structure.members[j]
                displacements = mem.nodal_displacements.values()
                max_vals = [max(l[0:2]) for l in displacements]
                min_vals = [min(l[0:2]) for l in displacements]
                max_val = max(max_vals)
                min_val = min(min_vals)
                abs_max = max(max_val, abs(min_val))
                return abs_max / self.delta_max - 1

            constr = NonLinearConstraint(con_fun,
                                         name='Displacement ' + str(11 + j),
                                         parent=self)
            constr.fea_required = True
            self.cons.append(constr)

    def create_objective(self):

        def objective(X):
            weight = 0
            for x, mem in zip(X, self.structure.members.values()):
                weight += self.rho * mem.A * mem.length
            return weight

        self.obj = objective


def draw_structure(structure):
    from pyautocad import Autocad, APoint
    acad = Autocad(create_if_not_exists=True)
    scale = min(structure.L, structure.H)
    text_size = scale / 40
    node_size = scale / 100

    for mem in structure.members.values():
        [x1, y1], [x2, y2] = mem.coordinates
        p1 = APoint(x1, y1)
        p2 = APoint(x2, y2)
        acad.model.AddLine(p1, p2)
        # Nodes
        acad.model.AddCircle(p1, node_size)
        acad.model.AddCircle(p2, node_size)

        p1.x += mem.length / 4
        p1.y = mem.shape(p1.x)
        acad.model.AddText(f'{mem.mem_id + 1}', p1, text_size)

    for pl in structure.point_loads.values():
        x1, y1 = pl.coordinate
        fx, fy, m = np.asarray(pl.v)

        p1 = APoint(x1, y1)
        p2 = APoint(x1 + fx / 5e2, y1 + fy / 5e2)
        acad.model.AddLine(p1, p2)
        p2.x += scale / 100
        p2.y += scale / 100
        acad.model.AddText(f'F: {abs(fy) / 1000 :.2f} kN', p2, text_size)


def draw_deflected(structure):
    from pyautocad import Autocad, APoint
    acad = Autocad(create_if_not_exists=True)
    for mem in structure.members.values():
        [x1, y1], [x2, y2] = mem.coordinates
        p1 = APoint(x1, y1)
        p2 = APoint(x2, y2)
        delta1 = list(mem.nodal_displacements.values())[0]
        delta2 = list(mem.nodal_displacements.values())[-1]
        dx1, dy1, dr1 = delta1
        dx2, dy2, dr2 = delta2

        p1.x += dx1
        p1.y += dy1

        p2.x += dx2
        p2.y += dy2

        acad.model.AddLine(p1, p2)
        acad.model.AddCircle(p1, 100)
        acad.model.AddCircle(p2, 100)

    for pl in structure.point_loads.values():
        x1, y1 = pl.coordinate
        fx, fy, m = np.asarray(pl.v) / 5e2
        p1 = APoint(x1, y1)
        p2 = APoint(x1 + fx, y1 + fy)
        acad.model.AddLine(p1, p2)


def plot_values(problem, loc=[22, 15], X_vals=[]):
    OBJ_VALS = np.zeros((len(THREE_BAR_AREAS_mm2), len(THREE_BAR_AREAS_mm2)))
    CON_VALS = np.zeros((len(THREE_BAR_AREAS_mm2), len(THREE_BAR_AREAS_mm2)))
    for i in range(len(THREE_BAR_AREAS_mm2)):
        for j in range(len(THREE_BAR_AREAS_mm2)):
            X = [j, i]
            obj_val = problem.obj(X)
            con_val = max(problem.eval_con(X))
            if con_val > 0:
                con_val = -1
            else:
                con_val = 1

            OBJ_VALS[i, j] = obj_val
            CON_VALS[i, j] = con_val

    OBJ_CON = CON_VALS * OBJ_VALS

    import matplotlib.pyplot as plt
    from matplotlib.colors import ListedColormap
    from matplotlib import cm

    N = 16
    p = 0.2
    top = cm.get_cmap('Reds', N)
    bottom = cm.get_cmap('Greens', N)

    newcolors = np.vstack((top(np.linspace(p, 1, N)),
                           bottom(np.linspace(p, 1, N))))
    newcmp = ListedColormap(newcolors, name='OrangeBlue')

    vmax = np.max(OBJ_VALS)
    vmin = -vmax
    plt.matshow(OBJ_VALS * CON_VALS, cmap=newcmp, vmin=vmin, vmax=vmax,
                interpolation='nearest', aspect='equal')
    if len(X_vals):
        for i, loc in enumerate(X_vals):
            if i > 0:
                plt.plot([prev_loc[0], loc[0]],
                         [prev_loc[1], loc[1]], 'k')
            prev_loc = loc
    else:
        plt.text(loc[0] - 0.5, loc[1] + 0.5, "X")
    plt.colorbar()
    plt.xticks(np.arange(0, len(THREE_BAR_AREAS_mm2)))
    plt.yticks(np.arange(0, len(THREE_BAR_AREAS_mm2)))
    plt.gca().set_xticks([x - 0.5 for x in plt.gca().get_xticks()][1:],
                         minor='true')
    plt.gca().set_yticks([y - 0.5 for y in plt.gca().get_yticks()][1:],
                         minor='true')
    plt.grid(which='minor')
    plt.show()

    # X = np.arange(0, len(THREE_BAR_AREAS_mm2))
    # Y = np.arange(0, len(THREE_BAR_AREAS_mm2))
    # Z = CON_VALS
    # CON = plt.contour(X, Y, Z, levels=[1-1e-6])
    # plt.clabel(CON, inline=True)
    # CS = plt.contour(X, Y, OBJ_VALS, levels=20, cmap='Greens_r')
    # plt.clabel(CS, inline=True)
    # plt.grid(True)


def plot_one_hot(areas=TEN_BAR_AREAS_mm2, point=ten_bar_DLM):
    one_hot = np.eye(len(areas))[point]
    plt.matshow(one_hot, cmap="Greys_r")
    plt.xticks(np.arange(0, len(areas)))
    plt.yticks(np.arange(0, len(point)))
    plt.gca().set_xticks([x - 0.5 for x in plt.gca().get_xticks()][1:],
                         minor='true')
    plt.gca().set_yticks([y - 0.5 for y in plt.gca().get_yticks()][1:],
                         minor='true')
    plt.grid(False)
    plt.show()

def plot_convex_hull():
    from scipy.spatial import ConvexHull, convex_hull_plot_2d
    points = []
    for A, Iy in zip(RHS_A, RHS_Iy):
        points.append([A, Iy])
    points = np.asarray(points)
    hull = ConvexHull(points)
    import matplotlib.pyplot as plt

    plt.plot(points[:, 0], points[:, 1], 'o')
    for simplex in hull.simplices:
        plt.plot(points[simplex, 0], points[simplex, 1], 'k-')

    plt.plot(points[hull.vertices, 0], points[hull.vertices, 1], 'r--', lw=2)
    plt.plot(points[hull.vertices[0], 0], points[hull.vertices[0], 1], 'ro')
    plt.show()

if __name__ == '__main__':
    #problem = TenBarTruss(prob_type='continuous')
    #solver = SLSQP()
    #solver.solve(problem, maxiter=100)

    problem = ThreeBarTruss("continuous")
    #problem = FifteenBarTruss(prob_type='continuous')
    x0 = [var.ub for var in problem.vars]
    solver = SLP(move_limits=[0.25, 0.25], gamma=1e-3)
    fopt, xopt = solver.solve(problem, x0=x0, maxiter=100, log=False, verb=True, plot=False)
    problem(xopt)

    # solver = GA()
    # solver.solve(problem, maxiter=100 )
    # solver = GA(popsize=15)
    # solver.solve(problem, maxiter=50)
    # X, fopt, X_vals = solver.solve(problem,maxiter=5000, maxtime=3, subset_size=-1)
    # plot_values(problem, X_vals=X_vals)
    # problem([20, 13])
    # plot_values(problem)
    # solver.solve(problem, maxiter=5000)

    #plot_convex_hull()
    #plot_one_hot()
    # draw_structure(problem.structure)
    # draw_deflected(problem.structure)



# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Timber member

@author Viktor Haimi
"""
from math import log, tan, pi, sqrt, isclose, sin, cos, radians
import numpy as np

import metku.frame2d as frame2d
from metku.materials.timber_data import Timber, T
from metku.eurocodes.en1995.en1995_1_1 import kmod, kdef, k_to_d, get_gammaM
from metku.eurocodes.en1995.constants import k_m
from metku.sections.timber.timber_section import *
from metku.framefem.framefem import FEMNode
from metku.framefem.elements import EBBeam, EBSemiRigidBeam, Rod
from metku.sections.steel.ISection import HEA

PREC = 5

import sys

if sys.version_info[0] == 3 and sys.version_info[1] < 9:
    from typing import List, Tuple, Dict
    list_array = List[np.ndarray]
    #list_float_FEMNode = List[[float,FEMNode]]
    list_float_FEMNode = Tuple[float,FEMNode]
    list_out = List[float]
    llist_out = List[List[float]]
    lllist_out = List[List[List[float]]]
    
    dict_str_float = Dict[str, float]
    
    tuple_out = Tuple
else:
    list_array = list[np.ndarray]
    list_float_FEMNode = list[float,FEMNode]
    list_out = list[float]
    llist_out = list[list[float]]
    lllist_out = list[list[list[float]]]
    
    dict_str_float = dict[str: float]
    
    tuple_out = tuple

class TimberMember:
    def __init__(self, profile: (TimberSection, TaperedSection), material: Timber, length: float, ldc: str, sc: int,
                 mtype: str = 'beam', varnished: bool = False, Sj1: int = np.inf, Sj2: int = np.inf, nodes: dict = None,
                 parent=None, lateral_support_y: list = None, lateral_support_z: list = None,
                 edge_load: str = 'compression', beta: (list, float) = None, k_vol: float=None):
        """
        Initialization method for timber member

        @param profile: poikkileikkaus luokkaa TimberSection
        @param material: sauvan materiaali muodossa: C14..D70 / GL24c..GL32h
        @param length: sauvan pituus [mm]
        @param ldc: Load duration class. options: 'perm' = permanent, 'lt' = long-term, 'mt' = medium-term,
                    'st' = short-term, 'inst' = instantaneous
        @param sc: Service class.
        @param mtype: member type: 'column', 'beam'
        @param varnished: lakkaus
        @param Sj1: Sauvan alkupään jäykkyys
        @param Sj2: Sauvan loppupään jäykkyys
        @param nodes:
        @param parent: TimberFrameMember
        @param lateral_support_y: locations of the supports in global y direction (This has nothing to do with FEM)
        @param lateral_support_z: locations of the supports in z direction (This has nothing to do with FEM)
        @param edge_load: 'compression' or 'tension'
        @param beta: Eulerin nurjahduskerroin
        @param k_vol: pakotettu k_vol
        """

        self.material = material
        self.ldc = ldc
        self.sc = sc
        self.varnished = varnished
        self.length = length

        self.type = mtype
        self.Sj1 = Sj1
        self.Sj2 = Sj2
        self.beta = beta
        self.perm_k_vol = k_vol

        self.ned = []
        self.myed = []
        self.mzed = []
        self.vyed = []
        self.vzed = []
        self.ted = []
        self.loc = []
        self.utilization = []

        self.nodes = nodes
        self.parent = parent

        self.lateral_support_y = lateral_support_y
        self.lateral_support_z = lateral_support_z
        self.L_c = np.asarray([np.asarray([]),
                               np.asarray([])], dtype=object)

        self.edge_load = edge_load

        self.kmod = kmod(self.material.type, self.sc, self.ldc)

        self.section = profile

        self.R = 0

        self.gammaM = get_gammaM(self.material.type)

    @property
    def k_L(self):
        # SFS-EN 1995-1-1 (3.4)
        k_L = 1.0
        if self.material.type == 'lvl':
            if self.NEd > 0 and self.length != 3000:
                k_L = min(1.1, (3000 / self.length) ** (self.material.s / 2))
        return k_L

    def get_kdef(self):
        return kdef(self.material.type, self.sc)

    @property
    def R(self) -> (int, float):
        return self.section.R

    @R.setter
    def R(self, val: (int, float)):
        self.section.R = val

    @property
    def fmd(self) -> float:
        if self.material.type == 'lvl':
            if self.material.direction == 'flat':
                return self.section.k_h * k_to_d(self.material.fm0flatk, self.kmod, self.gammaM)
        return self.section.k_h * k_to_d(self.material.fmk, self.kmod, self.gammaM)

    # ilman k_h kerrointa
    @property
    def unmodified_fmd(self) -> float:
        return k_to_d(self.material.fmk, self.kmod, self.gammaM)

    @property
    def ft0d(self) -> float:
        if self.material.type == 'solid_timber' or self.material.type == 'glt':
            return self.section.k_h * k_to_d(self.material.ft0k, self.kmod, self.gammaM)
        elif self.material.type == 'lvl':
            return self.k_L * k_to_d(self.material.ft0k, self.kmod, self.gammaM)

    # ilman k_h kerrointa
    @property
    def unmodified_ft0d(self) -> float:
        return k_to_d(self.material.ft0k, self.kmod, self.gammaM)

    @property
    def ft90d(self) -> float:
        if self.material.type == 'lvl':
            if self.material.direction == 'edge':
                return k_to_d(self.material.ft90edgek, self.kmod, self.gammaM)
            raise Exception('lvl does not have ft90d, try ft90edged')
        return k_to_d(self.material.ft90k, self.kmod, self.gammaM)

    @property
    def fc0d(self) -> float:
        return k_to_d(self.material.fc0k, self.kmod, self.gammaM)

    @property
    def fc90d(self) -> float:
        if self.material.type == 'lvl':
            if self.material.direction == 'flat':
                return k_to_d(self.material.fc90flatk, self.kmod, self.gammaM)
            return k_to_d(self.material.fc90edgek, self.kmod, self.gammaM)
        return k_to_d(self.material.fc90k, self.kmod, self.gammaM)

    @property
    def fvd(self) -> float:
        return k_to_d(self.material.fvk, self.kmod, self.gammaM)

    @property
    def ft90edged(self) -> float:
        if self.material.type == 'glt' or self.material.type == 'solid_timber':
            raise Exception(f'{self.material.type} does not have ft90edged, try ft90d')
        return k_to_d(self.material.ft90edgek, self.kmod, self.gammaM)

    @property
    def fr0d(self) -> float:
        if self.material.type == 'glt' or self.material.type == 'solid_timber':
            raise Exception(f'{self.material.type} does not have fr0d')
        return k_to_d(self.material.fr0k, self.kmod, self.gammaM)

    @property
    def MEdY(self) -> float:
        return max(abs(np.array(self.myed)))

    @property
    def MEdZ(self) -> float:
        return max(abs(np.array(self.mzed)))

    @property
    def NEd(self) -> float:
        """ Maximum axial force (absolute value) """
        return max(abs(np.array(self.ned)))

    @property
    def A(self) -> (int, float):
        return self.section.get_A(0)

    def get_A(self, at: float) -> (int, float):
        return self.section.get_A(at)

    @property
    def Iy(self) -> float:
        return self.section.I[0]

    @Iy.setter
    def Iy(self, val: float):
        self.section.I[0] = val

    @property
    def Iz(self) -> float:
        return self.section.I[1]

    @Iz.setter
    def Iz(self, val: float):
        self.section.I[1] = val

    @property
    def E0mean(self) -> (int, float):
        return self.material.E0mean

    @E0mean.setter
    def E0mean(self, val: (int, float)):
        self.material.E0mean = val

    @property
    def h(self) -> float:
        """
        Property, returns member's cross-section's height
        """
        return self.section.get_H(0)

    @h.setter
    def h(self, val: (int, float)):
        """
        Sets cross-section's height to given value.
        """
        self.section.H = val

    @property
    def b(self) -> float:
        """
        Property, returns member's cross-section's breadth
        """
        return self.section.B

    @b.setter
    def b(self, val: (int, float)):
        """
        Sets cross-section's height to given value.
        """
        self.section.B = val

    def sigma_t0d(self, n: int) -> float:
        """
        EN 1995-1-1 6.1.2
        @param n: indeksi
        @return:
        """
        return self.ned[n] / self.section.get_A(self.loc[n])

    def sigma_c0d(self, n: int) -> float:
        """
        EN 1995-1-1 6.1.4
        @param n: indeksi
        @return:
        """
        return self.ned[n] / self.section.get_A(self.loc[n])

    def sigma_m_alpha_d(self) -> float:
        # SFS 1995-1-1 (6.37)
        list = [self.myed[self.loc.index(i)] / self.section.get_Wel(i)[0] for i in self.loc]
        min_val = min(list)
        max_val = max(list)
        abs_max = max(abs(min_val), abs(max_val))
        if abs_max == abs(min_val):
            return min_val
        else:
            return max_val

    def sigma_t90d(self, kaava6_55:bool=False) -> float:
        if kaava6_55:
            p = 0
            for load in self.parent.loads:
                if load.direction == 'y' and load.values[0] == load.values[1] and self.edge_load == 'compression':
                    p = load.values[0]
            # SFS 1995-1-1 (6.55)
            return self.k_p() * self.M_apd() / self.W_ap()[0] - 0.6 * (p / self.section.B)
        # SFS 1995-1-1 (6.54)
        return self.k_p() * self.M_apd() / self.W_ap()[0]

    def M_apd(self) -> float:
        """
        harjavyöhykkeen momentti
        jos solmu löytyy keskeltä niin käytetään sitä, muuten lasketaan kahden viereisen solmun keskiarvo
        @return:
        """
        try:
            return self.myed[self.loc.index(self.section.xap)]
        except:
            minusloc = None
            for loc in self.loc:
                if minusloc is None:
                    minusloc = loc
                    continue
                diff1 = self.section.xap - loc
                diff2 = self.section.xap - minusloc
                if diff2 > diff1 >= 0:
                    minusloc = loc
            plusloc = None
            for loc in self.loc[::-1]:
                if plusloc is None:
                    plusloc = loc
                    continue
                diff1 = loc - self.section.xap
                diff2 = plusloc - self.section.xap
                if diff2 > diff1 >= 0:
                    plusloc = loc
            return (self.myed[self.loc.index(minusloc)] + self.myed[self.loc.index(plusloc)]) / 2

    def W_ap(self) -> np.ndarray:
        """
        harjavyöhykkeen kimmoinen taivutusvastus
        @return:
        """
        return self.section.get_Wel(self.section.xap)

    def k_r(self) -> float:
        # SFS 1995-1-1 (6.49)
        # TODO Mahapalkki kaarevien ryhmään, vaatii mahapalkin uudelleen määrittämisen
        if isinstance(self.section, (HarjaPalkki, PulpettiPalkki, MahaPalkki)):
            return 1
        elif isinstance(self.section, (KaarevaHarjaPalkki, KaarevaPalkki)):
            if self.section.r_in * 1000 / self.section.t >= 240:
                return 1
            else:
                return 0.76 + 0.001 * (self.section.r_in * 1000 / self.section.t)
        else:
            raise NotImplementedError

    def V(self) -> float:
        # SFS 1995-1-1 (Kuva 6.9)
        # V <= 2/3 * koko palkin tilavuus
        if isinstance(self.section, (HarjaPalkki, MahaPalkki)):
            return self.section.B * np.power(self.section.hap, 2)
        elif isinstance(self.section, KaarevaHarjaPalkki):

            # l1 = self.section.h0 * cos(self.alfa())
            # l2 = (self.section.up_y1 - self.section.dwn_y1) / cos(self.alfa())
            # l3 = self.section.up_x1 / cos(self.alfa())
            # V_leg = self.section.B * l3 * (l1 + l2) * 0.5
            # r = self.V_tot() - 2 * V_leg

            from scipy import integrate
            up = integrate.quad(self.section.f1, 0, self.section.up_x1)
            dwn = integrate.quad(self.section.f2, 0, self.section.up_x1)
            intlegs = (up[0] - dwn[0]) * 2 * self.section.B

            r = self.V_tot() - intlegs

            if r < 0:
                print(f'harja-alueen tilavuudessa laskuvirhe, epätarkkojen oletusten vuoksi')
                # if intlegs < 0:
                #     print(self.section.intf1)
                #     print(self.section.intf2)
                #     print(self.section.intf3)
                #     print(self.section.intf4)
                #     print(self.section.intf6)
            return r
        elif isinstance(self.section, KaarevaPalkki):
            return self.section.B * self.section.hap * (self.section.r_in + self.section.hap * 0.5) * ((self.alfa() + self.section.beta) * 0.5)
        else:
            raise NotImplementedError

    def V_tot(self) -> float:
        if isinstance(self.section, (HarjaPalkki, MahaPalkki)):
            return self.section.B * self.length * (self.section.hap - self.length * np.tan(self.alfa()) / 4)
        elif isinstance(self.section, (KaarevaPalkki, KaarevaHarjaPalkki)):
            return self.section.A_face() * self.section.B
        elif isinstance(self.section, PulpettiPalkki):
            return (self.section.H0 + self.section.Hap) / 2 * self.length * self.section.B
        else:
            return self.section.A * self.length

    def k_vol(self) -> float:
        # SFS 1995-1-1 (6.51)
        if self.perm_k_vol is not None:
            return self.perm_k_vol
        if self.material.type == 'solid timber':
            return 1.0
        v = self.V()
        if v > (2 / 3) * self.V_tot():
            v = (2 / 3) * self.V_tot()
        k_vol = np.power(1e7 / v, 0.2)
        return k_vol

    def k_dis(self) -> float:
        # SFS 1995-1-1 (6.52)
        if isinstance(self.section, (HarjaPalkki, PulpettiPalkki, MahaPalkki)):
            return 1.4
        elif isinstance(self.section, (KaarevaHarjaPalkki, KaarevaPalkki)):
            return 1.7

    def k_p(self) -> float:
        # SFS 1995-1-1 (6.56)
        if isinstance(self.section, (HarjaPalkki, MahaPalkki)):
            return 0.2 * np.tan(self.alfa())
        elif isinstance(self.section, KaarevaPalkki):
            r = self.section.r_in + 0.5 * self.section.hap
            return 0.25 * (self.section.hap / r)
        elif isinstance(self.section, KaarevaHarjaPalkki):
            k5 = 0.2 * np.tan(self.alfa())
            k6 = 0.25 - 1.5 * np.tan(self.alfa()) + 2.6 * np.tan(self.alfa()) ** 2
            k7 = 2.1 * np.tan(self.alfa()) - 4 * np.tan(self.alfa()) ** 2
            r = self.section.r_in + 0.5 * self.section.hap
            return k5 + k6 * (self.section.hap / r) + k7 * (self.section.hap / r) ** 2

    def k_l(self) -> float:
        # SFS 1995-1-1 (6.43)
        k1 = 1 + 1.4 * np.tan(self.alfa()) + 5.4 * np.tan(self.alfa()) ** 2
        if isinstance(self.section, (HarjaPalkki, MahaPalkki)):
            return k1
        elif isinstance(self.section, KaarevaPalkki):
            r = self.section.r_in + 0.5 * self.section.hap
            return 1 + 0.35 * (self.section.hap / r) + 0.6 * (self.section.hap / r) ** 2
        elif isinstance(self.section, KaarevaHarjaPalkki):
            r = self.section.r_in + 0.5 * self.section.hap
            k2 = 0.35 * - 8 * np.tan(self.alfa())
            k3 = 0.6 + 8.3 * np.tan(self.alfa()) - 7.8 * np.tan(self.alfa()) ** 2
            k4 = 6 * np.tan(self.alfa()) ** 2
            return k1 + k2 * (self.section.hap / r) + k3 * (
                    self.section.hap / r) ** 2 + k4 * (self.section.hap / r) ** 3

    def k_m_alpha(self) -> float:
        # SFS 1995-1-1 (6.39) ja (6.40)
        sigma_malphad = self.sigma_m_alpha_d()
        if not self.section.flipped:
            if sigma_malphad >= 0:
                kma = np.power(1 + np.power(np.tan(self.alfa()) * (4 / 3) * self.fmd / self.fvd, 2) +
                               np.power(np.power(np.tan(self.alfa()), 2) * self.fmd / self.fc90d, 2), -1 / 2)
            else:
                kma = np.power(1 + np.power(np.tan(self.alfa()) * (2 / 3) * self.fmd / self.fvd, 2) +
                               np.power(np.power(np.tan(self.alfa()), 2) * self.fmd / self.fc90d, 2), -1 / 2)
        else:
            if sigma_malphad <= 0:
                kma = np.power(1 + np.power(np.tan(self.alfa()) * (4 / 3) * self.fmd / self.fvd, 2) +
                               np.power(np.power(np.tan(self.alfa()), 2) * self.fmd / self.fc90d, 2), -1 / 2)
            else:
                kma = np.power(1 + np.power(np.tan(self.alfa()) * (2 / 3) * self.fmd / self.fvd, 2) +
                               np.power(np.power(np.tan(self.alfa()), 2) * self.fmd / self.fc90d, 2), -1 / 2)
        return kma

    def alfa(self) -> float:
        """
        Palkin yläpinnan kaltevuuskulma
        @return: radiaaneina
        """
        if isinstance(self.section, PulpettiPalkki):
            a = np.arctan((self.section.hap - self.section.h0) / self.length)
        elif isinstance(self.section, HarjaPalkki):
            a = np.arctan((self.section.hap - self.section.h0) / (self.length / 2))
        elif isinstance(self.section, (KaarevaHarjaPalkki, KaarevaPalkki, MahaPalkki)):
            a = self.section.alpha
        else:
            a = 0
        return a

    def sigma_c90d(self, node: FEMNode) -> (list_out, float):
        """
        SFS 1995-1-1 6.1.5
        Rajoitusehdoissa pitää olla määritettynä H > B, muuten ohjelma asettaa pilarin "väärinpäin"

        @param node: node where to check sigma c90d
        @return:     0 if there is no other member on this node,
                     float if there is only one other member on this node
                     list of floats if there is more than one members on this node
        """

        if len(node.parents) == 1:
            return 0

        other_members = list(filter(lambda v: v.member != self if True else False, node.parents.copy()))

        if len(other_members) == 0:
            return 0

        def get_sigma(om):
            other_locs = om.member.loc
            other_idx = None
            for i in range(om.member.nsect()):
                if isclose(node.x, om.to_global(other_locs[i])[0], rel_tol=10) and isclose(node.y,
                                                                                           om.to_global(other_locs[i])[
                                                                                               1],
                                                                                           rel_tol=10):
                    other_idx = i

            own_idx = None
            for i in range(self.nsect()):
                if isclose(node.x, self.parent.to_global(self.loc[i])[0], rel_tol=10) and isclose(node.y,
                                                                                                  self.parent.to_global(
                                                                                                      self.loc[i])[1],
                                                                                                  rel_tol=10):
                    own_idx = i

            if om.member.ned[other_idx] >= 0:
                return 0

            if own_idx == 0 or own_idx == self.loc[-1]:
                # Nolla edustaa a mittaa. RIL 205-1-2009 kohta 6.1.5
                # Rajoitusehdoissa pitää olla määritettynä H > B, muuten ohjelma asettaa pilarin "väärinpäin"
                lc90ef = om.cross_section.H + min(30, 0, om.cross_section.H) + min(30, om.cross_section.H)
            else:
                lc90ef = om.cross_section.H + min(30, om.cross_section.H) + min(30, om.cross_section.H)

            # SFS-EN 1995-1-1 § 6.1.5
            Aef = lc90ef * min(om.cross_section.B, self.section.B)
            sigma = om.member.ned[other_idx] / Aef
            return sigma

        if len(other_members) == 1:
            om = other_members[0]
            return get_sigma(om)
        else:
            return [get_sigma(om) for om in other_members]

    # def kc_perpendicular(self, l: (int, float), a: (int, float), l1: (int, float)) -> float:
    #     """
    #     Tukipainekerroin: RIL 205-1-2009 B.5.2a,
    #     @param l:   sen sauvan kosketuspinnan pituus, joka rasitetaan syysuuntaa vastaan esim. pilarin päällä oleva palkki
    #     @param a:   kosketuspinnan ulkopuolinen, puristetun palkin päätyyn asti ulottuva mitta, tai seuraavaan puristuspintaan
    #     @param l1:  puristuspintojen lähin välimatka toisiinsa nähden
    #     @return:    tukipainekerroin
    #     """
    #     self.update_kc90(l1, l)
    #     lc90ef = l + min(30, a, l) + min(30, l, l1 / 2)
    #     return lc90ef / l * self.k_c90

    def update_kc90(self, l1: (int, float), l: (int, float)):
        """
        Materiaalikohtainen, tukipainekertoimen määritykseen tarkoitettu kerroin: SFS-EN 1995-1-1 § 6.1.5

        Rajataan SFS-EN 1995-1-1 § 6.1.5 Kuva 6.2 b kohtaan

        @param l1:  kosketuspintojen reunojen välinen etäisyys
        @param l: Toisen palkin poikkileikkauksen korkeus
        @return:
        """
        self.k_c90 = 1.0
        if self.section.H * 2 <= l1:
            if self.material.type == 'solid_timber':
                self.k_c90 = 2.5
            elif self.material.type == 'glt' and l <= 400:
                self.k_c90 = 1.75

    def sigma_md(self, n: int) -> list_out:
        """
        EN-1995-1-1 6.1.6
        @param n: indeksi
        @return:
        """
        Wel = self.section.get_Wel(self.loc[n])
        r = [self.myed[n] / Wel[0],
             self.mzed[n] / Wel[1]]
        return r

    def get_max_sigma_per_segment(self) -> lllist_out:
        """
        hakee suurimman sigma myd ja mzd per sivuttaistuetun palkin jokaiselle tukemattomalle palkin välille.
        @return:
        """
        supp_y = self.lateral_support_y.copy()
        supp_z = self.lateral_support_z.copy()
        supp_y.append(0.0)
        supp_y.append(1.0)
        supp_z.append(0.0)
        supp_z.append(1.0)
        supp_y.sort()
        supp_z.sort()

        sig_md_list = [self.sigma_md(i) for i in range(self.nsect())]

        iy = []
        iz = []
        for i in range(len(supp_y) - 1):
            sig_md_in_seg = []
            locs = []
            for loc in self.loc:
                if supp_y[i] <= loc <= supp_y[i + 1]:
                    sig_md_in_seg.append(sig_md_list[self.loc.index(loc)])
                    locs.append(loc)
            iy.append(max(sig_md_in_seg, key=lambda x: abs(x[0])))

        for i in range(len(supp_z) - 1):
            sig_md_in_seg = []
            locs = []
            for loc in self.loc:
                if supp_z[i] <= loc <= supp_z[i + 1]:
                    sig_md_in_seg.append(sig_md_list[self.loc.index(loc)])
                    locs.append(loc)
            iz.append(max(sig_md_in_seg, key=lambda x: abs(x[0])))
        return [iy, iz]

    def tau_d(self, n: int) -> list_out:
        """
        Poikkileikkauksessa vaikuttava työntövoima muutetaan leikkausjännitykseski Jourawskin kaavaa hyväksikäyttäen
        @return: palauttaa suorakaidepoikkileikkauksen leikkausjännityksen arvon, ottaa huomioon k_cr kertoimen
        """
        k_cr = 1.0
        if self.material.type == 'solid_timber' or self.material.type == 'glt' or self.sc == 1:
            if not self.varnished:
                k_cr = 0.67

        return [(3 / 2 * self.vyed[n]) / (self.section.get_H(self.loc[n]) * self.section.B * k_cr),
                (3 / 2 * self.vzed[n]) / (self.section.get_H(self.loc[n]) * self.section.B * k_cr)]

    def tau_tord(self, n: int) -> float:
        """
        EN-1995-1-1 6-1-8
        Alpha and beta are based on the theoretical tables given in: T. Salmi 2000, Lujuusopin perusteet, p. 258
        """
        section_height_to_width_ratio = self.section.get_H(self.loc[n]) / self.section.B
        if section_height_to_width_ratio < 1:
            section_height_to_width_ratio = 1 / section_height_to_width_ratio
        if 0.95 < section_height_to_width_ratio < 1.05:
            beta = 0.208
        elif 1.45 < section_height_to_width_ratio < 1.55:
            beta = 0.231
        elif 1.95 < section_height_to_width_ratio < 2.05:
            beta = 0.246
        elif 2.45 < section_height_to_width_ratio < 2.55:
            beta = 0.258
        elif 2.95 < section_height_to_width_ratio < 3.05:
            beta = 0.267
        elif 3.95 < section_height_to_width_ratio < 4.05:
            beta = 0.282
        elif 4.95 < section_height_to_width_ratio < 5.05:
            beta = 0.299
        elif 5.95 < section_height_to_width_ratio < 6.05:
            beta = 0.307
        elif 7.95 < section_height_to_width_ratio < 8.05:
            beta = 0.313
        elif 9 < section_height_to_width_ratio < np.inf:
            beta = 0.333
        else:
            beta = -0.0007 * section_height_to_width_ratio ** 2 + 0.0193 * section_height_to_width_ratio + 0.1923
        Wv = beta * self.section.get_H(self.loc[n]) * self.section.B ** 2
        return self.ted[n] / Wv

    def I_tor(self, loc: float) -> float:
        """
        @return: torsional constant for rectangular cross sections
        """
        ratio = self.section.get_H(loc) / self.section.B
        if ratio < 1:
            ratio = 1 / ratio
        if 1 < ratio < 1.05:
            alpha = 0.141
        elif 1.45 < ratio < 1.55:
            alpha = 0.196
        elif 1.95 < ratio < 2.05:
            alpha = 0.229
        elif 2.45 < ratio < 2.55:
            alpha = 0.249
        elif 2.95 < ratio < 3.05:
            alpha = 0.263
        elif 3.95 < ratio < 4.05:
            alpha = 0.281
        elif 4.95 < ratio < 5.05:
            alpha = 0.299
        elif 5.95 < ratio < 6.05:
            alpha = 0.307
        elif 7.95 < ratio < 8.05:
            alpha = 0.313
        elif 9 < ratio < np.inf:
            alpha = 0.333
        else:
            alpha = -0.0022 * ratio ** 2 + 0.425 * ratio + 0.1116

        return alpha * self.section.get_H(loc) * self.section.B ** 3

    def k_shape(self, n: int) -> float:
        """
        SFS EN 1995-1-1 §6.1.8(6.15)
        @return: cross section factor for rectangular cross sections (circular sections not included)
        """
        return min(1.3, 1 + 0.05 * self.section.get_H(self.loc[n]) / self.section.B)

    def beta_c(self) -> float:
        """
        EN 1995-1-1 6.3.2 (6.29)
        @return:
        """
        if self.material.type == 'solid_timber':
            return 0.2
        return 0.1

    def radius_of_gyration(self, get_locs: bool = False) -> (np.ndarray, list):
        """
        jäyhyyssäteen arvo etsitään suurimman taivutusjännityksen alueelta
        @param get_locs:
        @return:
        """
        supp_y = self.lateral_support_y.copy()
        supp_z = self.lateral_support_z.copy()
        supp_y.append(0.0)
        supp_y.append(1.0)
        supp_z.append(0.0)
        supp_z.append(1.0)
        supp_y.sort()
        supp_z.sort()

        sig_md_list = [self.sigma_md(i) for i in range(self.nsect())]
        iy = []
        iz = []
        for i in range(len(supp_y) - 1):
            sig_md_in_seg = []
            locs = []
            for loc in self.loc:
                if supp_y[i] <= loc <= supp_y[i + 1]:
                    sig_md_in_seg.append(sig_md_list[self.loc.index(loc)])
                    locs.append(loc)
            iy.append(locs[sig_md_in_seg.index(max(sig_md_in_seg))])

        for i in range(len(supp_z) - 1):
            sig_md_in_seg = []
            locs = []
            for loc in self.loc:
                if supp_z[i] <= loc <= supp_z[i + 1]:
                    sig_md_in_seg.append(sig_md_list[self.loc.index(loc)])
                    locs.append(loc)
            iz.append(locs[sig_md_in_seg.index(max(sig_md_in_seg))])

        y = [sqrt(self.section.get_I(i)[0] / self.section.get_A(i)) for i in iy]
        z = [sqrt(self.section.get_I(i)[1] / self.section.get_A(i)) for i in iz]
        if get_locs:
            return [np.asarray([np.asarray(y),
                                np.asarray(z)], dtype=object), [iy, iz]]
        else:
            return np.asarray([np.asarray(y),
                               np.asarray(z)], dtype=object)

    def Ln(self) -> (llist_out, np.ndarray):
        """
        Eulerin teoriaan perustuva nurjahduspituus
        @return:
        """
        # Beta arvot saatu Liimapuukäsikirja Osa 3 sivu 3-35
        # TODO tarkista betan arvot palkille ja pilarille erikseen Lpkk3 (3-39)
        beta = [[], []]

        def form_beta(lc: list, index: int, beam=False):
            startsupp = None
            endsupp = None
            for supp in self.parent.frame2d.supports.values():
                if list(self.parent.n1.coord) == supp.coordinate:
                    startsupp = supp
                elif list(self.parent.n2.coord) == supp.coordinate:
                    endsupp = supp

            if len(lc) == 1:
                if beam:
                    if self.Sj1 < 1e3 and self.Sj2 < 1e3:
                        beta[index] = 1
                    elif (self.Sj1 >= 1e15 and self.Sj2 <= 1e3) or (self.Sj2 >= 1e15 and self.Sj1 <= 1e3):
                        beta[index] = 0.9
                    elif self.Sj1 >= 1e15 and self.Sj2 >= 1e15:
                        beta[index] = 0.8
                    else:
                        beta[index] = 0.85

                elif startsupp is not None and endsupp is not None:
                    if isinstance(startsupp, frame2d.FixedSupport) and isinstance(endsupp, frame2d.FixedSupport):
                        beta[index] = 0.7
                    elif isinstance(startsupp, frame2d.XYHingedSupport) and isinstance(endsupp, frame2d.XYHingedSupport):
                        beta[index] = 1
                    elif (isinstance(startsupp, frame2d.FixedSupport) or isinstance(endsupp, frame2d.FixedSupport) and
                          isinstance(startsupp, frame2d.XYHingedSupport) or isinstance(endsupp, frame2d.XYHingedSupport)):
                        beta[index] = 0.85
                    elif isinstance(startsupp, frame2d.YHingedSupport) or isinstance(endsupp, frame2d.YHingedSupport):
                        if isinstance(startsupp, frame2d.FixedSupport) or isinstance(endsupp, frame2d.FixedSupport):
                            beta[index] = 2.2
                        elif isinstance(startsupp, frame2d.XYHingedSupport) or isinstance(endsupp, frame2d.XYHingedSupport):
                            beta[index] = 2
                    elif isinstance(startsupp, frame2d.XHingedSupport) or isinstance(endsupp, frame2d.XHingedSupport):
                        if isinstance(startsupp, frame2d.FixedSupport) or isinstance(endsupp, frame2d.FixedSupport):
                            beta[index] = 0.85
                        elif isinstance(startsupp, frame2d.XYHingedSupport) or isinstance(endsupp, frame2d.XYHingedSupport):
                            beta[index] = 1
                elif isinstance(startsupp, frame2d.FixedSupport) or isinstance(endsupp, frame2d.FixedSupport):
                    beta[index] = 2.2
                elif isinstance(startsupp, frame2d.XYHingedSupport) or isinstance(endsupp, frame2d.XYHingedSupport):
                    beta[index] = 2

            else:
                # jos sauvat on jaettu useampaan osaan/segmenttiin
                for i in range(len(lc)):
                    # palkin segmentin z suunnassa on kiinteä beta=1 nurjahduskerroin
                    if beam and index == 1:
                        beta[index].append(1)
                    # sauvan alkupäästä ensimmäinen segmentti beta=0.8, sitä seuraavat segmentit beta=1, jne.
                    # beta arvot annetaan tukien vapausasteiden ja sauvan päiden kiertymisjäykkyyksien mukaan y- ja z-suunnissa
                    elif i == 0:
                        if isinstance(startsupp, frame2d.FixedSupport) or self.Sj1 >= 1e15:
                            beta[index].append(0.8)
                        elif isinstance(startsupp, frame2d.XYHingedSupport) or self.Sj1 <= 1e3:
                            beta[index].append(1)
                        elif 1e3 < self.Sj1 < 1e15:
                            beta[index].append(0.85)
                        else:
                            beta[index].append(1)
                    elif i == len(lc) - 1:
                        # sauvan alkupäästä viimeisessä segmentissä beta=0.8, sitä seuraavat segmentit beta=1, jne.
                        if isinstance(endsupp, frame2d.FixedSupport) or self.Sj2 >= 1e15:
                            beta[index].append(0.8)
                        elif isinstance(endsupp, frame2d.XYHingedSupport) or self.Sj2 <= 1e3:
                            beta[index].append(1)
                        elif 1e3 < self.Sj2 < 1e15:
                            beta[index].append(0.85)
                        else:
                            beta[index].append(1)
                    else:
                        # sauvojen välisegmenteille lokaalissa z-suunnassa beta=1, mutta y-suunnassa beta=0.9
                        if index == 0:
                            beta[index].append(0.9)
                        else:
                            beta[index].append(1)

        if self.beta is not None:
            if isinstance(self.beta, (int, float)):
                return np.asarray([self.beta * self.L_c[0], self.beta * self.L_c[1]], dtype=object)
            elif isinstance(self.beta, list):
                return np.asarray([self.beta[0] * self.L_c[0], self.beta[1] * self.L_c[1]], dtype=object)
        if self.type == 'column':
            form_beta(self.L_c[0], 0)
            form_beta(self.L_c[1], 1)
            self.beta = beta
            return np.asarray([beta[0] * self.L_c[0], beta[1] * self.L_c[1]], dtype=object)
        elif self.type in ('rigid-beam', 'beam'):
            form_beta(self.L_c[0], 0, True)
            form_beta(self.L_c[1], 1, True)
            self.beta = beta
            return np.asarray([beta[0] * self.L_c[0], beta[1] * self.L_c[1]], dtype=object)

        if self.Sj1 >= 1e15 and self.Sj2 >= 1e15:
            beta = 0.7
        elif (self.Sj1 >= 1e15 and self.Sj2 <= 1e3) or (self.Sj2 >= 1e15 and self.Sj1 <= 1e3):
            beta = 0.85
        else:
            beta = 1
        return [beta * self.L_c[0], beta * self.L_c[1]]

    def lamda(self) -> list_array:
        """
        EN 1995-1-1 6.3.2
        @return:
        """
        i = self.radius_of_gyration()
        ln = self.Ln()
        return [ln[0] / i[0],
                ln[1] / i[1]]

    def lamda_rel(self) -> list_array:
        """
        EN 1995-1-1 6.3.2 (6.21, 6.22)
        @return:
        """
        lam = self.lamda()
        return [(lam[0] / pi) * sqrt(self.material.fc0k / self.material.E005),
                (lam[1] / pi) * sqrt(self.material.fc0k / self.material.E005)]

    def sigma(self) -> list_array:
        """
        sigma kriittinen
        @return:
        """
        lam = self.lamda()
        return [pi ** 2 * (self.material.E0mean / (lam[0] ** 2)),
                pi ** 2 * (self.material.E0mean / (lam[1] ** 2))]

    def k(self) -> list_array:
        """
        EN 1995-1-1 6.3.2 (6.27, 6.28)
        @return:
        """
        l_rel = self.lamda_rel()
        return [0.5 * (1 + self.beta_c() * (l_rel[0] - 0.3) + np.power(l_rel[0], 2)),
                0.5 * (1 + self.beta_c() * (l_rel[1] - 0.3) + np.power(l_rel[1], 2))]

    def k_c(self) -> list_array:
        """
        EN 1995-1-1 6.3.2 (6.25, 6.26)
        @return:
        """
        # SFS-EN 1995-1-1 § 6.3.2 (6.26)
        k = self.k()
        l_rel = self.lamda_rel()
        r = [np.power(k[0] + np.power(np.power(k[0], 2) - np.power(l_rel[0], 2), 0.5), -1),
             np.power(k[1] + np.power(np.power(k[1], 2) - np.power(l_rel[1], 2), 0.5), -1)]
        return r

    def l_ef(self, i: int, yz: int, loc: float) -> float:
        # SFS-EN 1995-1-1 § 6.3.3 (6.1)
        # TODO SFS 1995-1-1 s. 43 Taulukko 6.1
        # TODO Pistekuorma ja vakiomomentti tilanne puuttuu

        alpha = 0.9

        if self.edge_load == 'compression':
            return (self.L_c[yz][i] * alpha) + 2 * self.section.get_H(loc)
        elif self.edge_load == 'tension':
            return (self.L_c[yz][i] * alpha) - 0.5 * self.section.get_H(loc)

        return self.L_c[yz][i] * alpha

    def sigma_mcrit(self, i: int, yz: int, loc: float) -> float:
        """
        For member undergoing bending relative to it's stronger axis: SFS-EN 1995-1-1 § 6.3.3 (6.31)
        @return: Single value of critical stress during lateral buckling check
        """
        try:
            return pi * sqrt(self.material.E005 * self.section.get_I(loc)[1] * self.material.G005 * self.I_tor(loc)) / (
                        self.l_ef(i, yz, loc) * self.section.get_Wel(loc)[0])
        except:
            return ((0.78 * self.section.B ** 2) / (
                        self.section.get_H(loc) * self.l_ef(i, yz, loc))) * self.material.E005

    def lamda_relm(self, i: int, yz: int, loc: float) -> float:
        """
        SFS-EN 1995-1-1 § 6.3.3 (6.30)
        @return: Relative slenderness for bending problems
        """
        return np.sqrt(self.material.fmk / self.sigma_mcrit(i, yz, loc))

    def k_crit(self, i: int, yz: int, loc) -> float:
        # SFS-EN 1995-1-1 § 6.3.3 (6.34)
        if self.lamda_relm(i, yz, loc) <= 0.75:
            return 1
        elif 0.75 < self.lamda_relm(i, yz, loc) <= 1.4:
            return 1.56 - 0.75 * self.lamda_relm(i, yz, loc)
        else:
            return 1 / self.lamda_relm(i, yz, loc) ** 2

    def check_single_tapered_beam_bending_tension(self) -> float:
        # SFS 1995-1-1 (6.38)
        return self.sigma_m_alpha_d() / (self.k_m_alpha() * self.fmd)

    def check_tapered_beam_bending_tension(self) -> float:
        """
        EN 1995-1-1 6.4.3 (6.41) kuva 6.9
        @return: APXBS käyttöasteen
        """
        try:
            sig = self.k_l() * self.sigma_md(self.loc.index(self.section.xap))[0]
        except:
            minusloc = None
            for loc in self.loc:
                if minusloc is None:
                    minusloc = loc
                    continue
                diff1 = self.section.xap - loc
                diff2 = self.section.xap - minusloc
                if diff2 > diff1 >= 0:
                    minusloc = loc
            plusloc = None
            for loc in self.loc[::-1]:
                if plusloc is None:
                    plusloc = loc
                    continue
                diff1 = loc - self.section.xap
                diff2 = plusloc - self.section.xap
                if diff2 > diff1 >= 0:
                    plusloc = loc
            sig = self.k_l() * (
                    self.sigma_md(self.loc.index(minusloc))[0] + self.sigma_md(self.loc.index(plusloc))[
                0]) / 2

        return sig / (self.k_r() * self.fmd)

    def check_perpendicular_to_grain_tension(self) -> float:
        """
        EN 1995-1-1 6.4.3 (6.50)
        @return: APXPT käyttöasteen
        """
        return self.sigma_t90d() / (self.k_dis() * self.k_vol() * self.ft90d)

    def check_normal_force(self, n: int) -> float:
        # Normal force utilization ratio along the grain:           SFS-EN 1995-1-1 § 6.1.2 and 6.1.4
        if self.ned[n] > 0:
            UN = abs(self.sigma_t0d(n)) / self.ft0d
        else:
            UN = abs(self.sigma_t0d(n)) / self.fc0d
        r = UN
        return r

    # def check_perpendicular_tension(self):
    #     # Member tension perpendicular to the grain:                SFS-EN 1995-1-1 § 6.1.3
    #     # TODO syyn suuntaan vastaset voimat
    #     pass

    def check_perpendicular_compression(self, location: bool = False) -> (list_float_FEMNode, float):
        # Member compression perpendicular to the grain:            SFS-EN 1995-1-1 § 6.1.5

        # TODO ei laske oikein yhdessä pisteessä kahdelta suunnalta puristettua palkkia
        max_sigma = 0
        node = None
        l1 = None
        for n in self.nodes.values():
            n_sig = self.sigma_c90d(n)
            if isinstance(n_sig, list):
                n_sig = min(n_sig)
            if abs(n_sig) > abs(max_sigma):
                max_sigma = n_sig
                node = n
        if node:
            for n in self.nodes.values():
                if len(n.parents) == 1 or n == node:
                    continue

                other_members = list(filter(lambda v: v.member != self if True else False, n.parents.copy()))

                if len(other_members) == 0:
                    continue

                start_node = n.coord
                end_node = node.coord
                l11 = np.linalg.norm(end_node - start_node)
                if l1 is None:
                    l1 = l11
                elif l11 < l1:
                    l1 = l11
                self.update_kc90(l1, other_members[0].cross_section.get_H(0))

        if not hasattr(self, 'k_c90'):
            if location:
                return [0, node]
            return 0

        r = abs(max_sigma / (self.k_c90 * self.fc90d))
        if location:
            return [r, node]
        return r

    def check_shear_force(self, n: int, reduce: bool = True) -> float:
        """
        Laskee leikkausjännityksen käyttöasteen annetun indeksin pisteessä

        Jos pienennys on True niin mitoittavan leikkausvoiman pienennys simuloidaan RIL 205-1-2009
        kohdan 6.1.7 mukaisesti palkin päissä.
        Jos annettu indeksi on pienennysalueen sisällä niin etsitään indeksi joka on lähimpänä pienennysalueen reunaa,
        mutta silti pienennyksalueen sisällä

        @param n: indeksi
        @param reduce: pienennys
        @return: käyttöaste
        """
        # Shear force utilization ratio:                            SFS-EN 1995-1-1 § 6.1.7

        if reduce:
            if self.type in ("beam", "rigid-beam"):
                start = list(filter(lambda v: v.member != self if True else False, self.parent.n1.parents.copy()))
                if len(start) == 0:
                    other_h_start = 0
                else:
                    other_h_start = start[0].cross_section.H / 2
                sumH_start = (other_h_start + self.section.get_H(0)) / self.length
                loc = self.loc[n]
                if loc < sumH_start:
                    for i in range(len(self.loc)):
                        try:
                            if self.loc[n + i + 1] > sumH_start:
                                n = n + i
                                break
                        except:
                            pass

                end = list(filter(lambda v: v.member != self if True else False, self.parent.n2.parents.copy()))
                if len(end) == 0:
                    other_h_end = 0
                else:
                    other_h_end = end[0].cross_section.H / 2
                sumH_end = (other_h_end + self.section.get_H(0)) / self.length
                if loc > 1 - sumH_end:
                    for i in range(len(self.loc)):
                        try:
                            if self.loc[n - i - 1] < 1 - sumH_end:
                                n = n - i
                                break
                        except:
                            pass

        UVy = abs(self.tau_d(n)[0]) / self.fvd
        UVz = abs(self.tau_d(n)[1]) / self.fvd
        r = max(UVy, UVz)
        return r

    def check_bending_moment(self, n: int) -> float:
        # Bending moment utilization ratio:                         SFS-EN 1995-1-1 § 6.1.6
        UM1 = abs(self.sigma_md(n)[0] / self.fmd + k_m * self.sigma_md(n)[1] / self.fmd)
        UM2 = abs(self.sigma_md(n)[1] / self.fmd + k_m * self.sigma_md(n)[0] / self.fmd)
        r = max(UM1, UM2)
        return r

    def check_torsion(self, n: int) -> float:
        # Torsion utilization ratio:                                SFS-EN 1995-1-1 § 6.1.8
        UTx = self.tau_tord(n) / (self.k_shape(n) * self.fvd)
        r = UTx
        return r

    def check_bending_tension(self, n: int) -> float:
        #                                                           SFS-EN 1995-1-1 § 6.2.3
        # TODO z-suuntainen taivutusjännitys tarkistettava merkkien osalta (deplanaatiota ei ole otettu huomioon)
        UM1 = self.sigma_t0d(n) / self.ft0d + abs(self.sigma_md(n)[0]) / self.fmd + k_m * self.sigma_md(n)[1] / self.fmd
        UM2 = self.sigma_t0d(n) / self.ft0d + k_m * abs(self.sigma_md(n)[0]) / self.fmd + self.sigma_md(n)[1] / self.fmd
        r = max(UM1, UM2)
        return r

    def check_bending_compression(self, n: int) -> float:
        #                                                           SFS-EN 1995-1-1 § 6.2.4
        # TODO katso ylemää
        UM1 = (self.sigma_c0d(n) / self.fc0d) ** 2 + abs(self.sigma_md(n)[0]) / self.fmd + k_m * self.sigma_md(n)[1] / self.fmd
        UM2 = (self.sigma_c0d(n) / self.fc0d) ** 2 + k_m * abs(self.sigma_md(n)[0]) / self.fmd + self.sigma_md(n)[1] / self.fmd
        r = max(UM1, UM2)
        return r

    def check_buckling(self) -> list_out:
        # Utilization ratio against member buckling:                SFS-EN 1995-1-1 § 6.3.2 (6.23, 6.24))
        sigma_c0d_list = [[], []]
        sigma_myd_list = [[], []]
        sigma_mzd_list = [[], []]

        def form_max_sigma_list(supports: list_out, index: int):
            supp = supports.copy()
            supp.append(0.0)
            supp.append(1.0)
            supp.sort()
            for i in range(len(supp) - 1):
                max_sigma_c0d_list = []
                max_sigma_myd_list = []
                max_sigma_mzd_list = []
                for loc in self.loc:
                    if supp[i] <= loc <= supp[i + 1]:
                        max_sigma_c0d_list.append(self.sigma_c0d(self.loc.index(loc)))
                        max_sigma_myd_list.append(self.sigma_md(self.loc.index(loc))[0])
                        max_sigma_mzd_list.append(self.sigma_md(self.loc.index(loc))[1])
                max_sigma_c0d = min(max_sigma_c0d_list)
                max_sigma_myd = max(max_sigma_myd_list, key=abs)
                max_sigma_mzd = max(max_sigma_mzd_list, key=abs)

                sigma_c0d_list[index].append(max_sigma_c0d)
                sigma_myd_list[index].append(max_sigma_myd)
                sigma_mzd_list[index].append(max_sigma_mzd)

        form_max_sigma_list(self.lateral_support_y, 0)
        form_max_sigma_list(self.lateral_support_z, 1)

        sc0 = np.array(sigma_c0d_list[0])
        sc1 = np.array(sigma_c0d_list[1])

        my0 = np.array(sigma_myd_list[0])
        my1 = np.array(sigma_myd_list[1])
        mz0 = np.array(sigma_mzd_list[0])
        mz1 = np.array(sigma_mzd_list[1])

        UB = [(abs(sc0) / (self.k_c()[0] * self.fc0d) + abs(my0) / self.fmd + k_m * mz0 / self.fmd),
              (abs(sc1) / (self.k_c()[1] * self.fc0d) + k_m * abs(my1) / self.fmd + mz1 / self.fmd)]
        UB_fixed = [UB[0] * (sc0 < 0), UB[1] * (sc1 < 0)]

        return [max(UB_fixed[0]), max(UB_fixed[1])]

    def check_LT_buckling(self, get_full: bool = False) -> (float, dict_str_float):
        # Utilization ratio against member lateral buckling:        SFS-EN 1995-1-1 § 6.3.3 (3) and (6)
        ULT_list = []
        supp = self.lateral_support_z.copy()
        supp.append(0.0)
        supp.append(1.0)
        supp.sort()
        for i in range(len(supp) - 1):
            for loc in self.loc:
                if supp[i] <= loc <= supp[i + 1]:
                    if max(self.myed, key=abs) != 0 and abs(max(self.ned, key=abs)) <= 0.1:
                        ULT_list.append({'r': self.sigma_md(self.loc.index(loc))[0] / (self.k_crit(i, 1, loc) * self.fmd), 'kaava': 33, 'loc': loc, 'i': i})
                    else:
                        ULT_list.append({
                            'r': ((self.sigma_md(self.loc.index(loc))[0] / (self.k_crit(i, 1, loc) * self.fmd)) ** 2) +
                            abs(self.sigma_c0d(self.loc.index(loc))) / (self.k_c()[1][i] * self.fc0d), 'kaava': 35, 'loc': loc, 'i': i})

                # TODO sigma_cod positiivinen tai negatiivinen
        if get_full:
            return max(ULT_list, key=lambda x: x['r'])
        return max(ULT_list, key=lambda x: x['r'])['r']

    def bracing(self, n: int=None, ks=None) -> (tuple_out[float, float, float], tuple_out[float, float]):
        """
        # SFS EN 1995-1-1 9.2.5
        kaavat 9.34, 9.35, 9.36, 9.37

        @param n: kehien määrä
        @param ks: ks
        @return: if n is given returns C, Fd and qd, and if its not given returns C and Fd
        """

        # Z-suunnan stabiloivan tuen voima ja jousijäykkyys 1.muoto
        m = len(self.lateral_support_z) + 1
        a = self.parent.length / m
        if ks is None:
            ks = 2 + 2 * cos(radians(180 / m))
            #print(f'ks {ks}')
        Nd_list = []
        supp = self.lateral_support_z.copy()
        supp.append(0.0)
        supp.append(1.0)
        supp.sort()
        for i in range(len(supp) - 1):
            for loc in self.loc:
                if supp[i] <= loc <= supp[i + 1]:
                    Md = self.myed[self.loc.index(loc)]
                    Nd_temp = (1 - self.k_crit(i, 1, loc)) * (Md / self.section.get_H(loc))
                    Nd_list.append(Nd_temp)

        Nd = max(Nd_list)
        C = Nd / a
        if self.material.type == 'solid_timber':
            Fd = Nd / 50
        else:
            Fd = Nd / 80

        if n is not None:
            qd = min([1, sqrt(15/self.parent.length)]) * ((n * Nd) / (50 * self.parent.length))
            return C, Fd, qd
        return C, Fd

    def check_apex_bending_stress(self) -> float:
        # EN 1995-1-1 (6.41)
        sigma = self.k_l() * self.M_apd() / self.W_ap()[0]
        return sigma / (self.k_r() * self.fmd)

    def check_apex_perpendicular_tension(self) -> float:
        # EN 1995-1-1 (6.50)
        return self.sigma_t90d() / (self.k_dis() * self.k_vol() * self.ft90d)

    def check_apex_shear_perpendicular_tension_combined(self, kaava6_55:bool=False) -> float:
        # EN 1995-1-1 (6.53)
        #
        n = 0
        start = list(filter(lambda v: v.member != self if True else False, self.parent.n1.parents.copy()))
        if len(start) == 0:
            other_h_start = 0
        else:
            other_h_start = start[0].cross_section.H / 2
        sumH_start = (other_h_start + self.section.get_H(0)) / self.length
        loc = self.loc[n]
        if loc < sumH_start:
            for i in range(len(self.loc)):
                try:
                    if self.loc[n + i + 1] > sumH_start:
                        n = n + i
                        break
                except:
                    pass

        tau1 = self.tau_d(n)[1]
        n = self.nsect() - 1
        end = list(filter(lambda v: v.member != self if True else False, self.parent.n2.parents.copy()))
        if len(end) == 0:
            other_h_end = 0
        else:
            other_h_end = end[0].cross_section.H / 2
        sumH_end = (other_h_end + self.section.get_H(0)) / self.length
        if loc > 1 - sumH_end:
            for i in range(len(self.loc)):
                try:
                    if self.loc[n - i - 1] < 1 - sumH_end:
                        n = n - i
                        break
                except:
                    pass
        tau2 = self.tau_d(n)[1]
        tau = max(tau1, tau2)
        # if isinstance(self.section, (KaarevaHarjaPalkki, KaarevaPalkki)):
        #     closesttodownx1 = None
        #     closesttodownx2 = None
        #     for loc in self.loc:
        #         if closesttodownx1 is None:
        #             closesttodownx1 = loc
        #             continue
        #         if closesttodownx2 is None:
        #             closesttodownx2 = loc
        #             continue
        #         if abs((self.section.dwn_x1 / self.length) - loc) < abs(closesttodownx1 - loc):
        #             closesttodownx1 = loc
        #         if abs(self.section.dwn_x2 / self.length - loc) < abs(closesttodownx2 - loc):
        #             closesttodownx2 = loc
        #
        #     tau_d_dw_x1 = self.tau_d(self.loc.index(closesttodownx1))
        #     tau_d_dw_x2 = self.tau_d(self.loc.index(closesttodownx2))
        #
        #     tau = abs(min(tau_d_dw_x1[1], tau_d_dw_x2[1], key=abs))
        #
        # elif isinstance(self.section, (HarjaPalkki, MahaPalkki)):
        #     closesttoleft = None
        #     closesttoright = None
        #     for loc in self.loc:
        #         if closesttoleft is None:
        #             closesttoleft = loc
        #             continue
        #         if closesttoright is None:
        #             closesttoright = loc
        #             continue
        #         if abs(self.section.xap - (self.section.hap / 2 / self.length) - loc) < abs(closesttoleft - loc):
        #             closesttoleft = loc
        #         if abs(self.section.xap + (self.section.hap / 2 / self.length) - loc) < abs(closesttoright - loc):
        #             closesttoright = loc
        #
        #     tau_d_dw_x1 = self.tau_d(self.loc.index(closesttoleft))
        #     tau_d_dw_x2 = self.tau_d(self.loc.index(closesttoright))
        #
        #     tau = abs(min(tau_d_dw_x1[1], tau_d_dw_x2[1], key=abs))
        # else:
        #     raise NotImplementedError
        return (tau / self.fvd) + self.sigma_t90d(kaava6_55) / (self.k_dis() * self.k_vol() * self.ft90d)

    def add_section(self, ned: float = 0.0, myed: float = 0.0, mzed: float = 0.0, vyed: float = 0.0, vzed: float = 0.0,
                    ted: float = 0.0, loc: float = 0.0):
        """ Adds new section with internal forces and their location """

        self.ned.append(ned)
        self.myed.append(myed)
        self.mzed.append(mzed)
        self.vyed.append(vyed)
        self.vzed.append(vzed)
        self.ted.append(ted)
        self.loc.append(loc)
        self.utilization.append({})

    def clear_sections(self):
        self.ned.clear()
        self.myed.clear()
        self.mzed.clear()
        self.vyed.clear()
        self.vzed.clear()
        self.ted.clear()
        self.loc.clear()
        self.utilization.clear()

    def check_section(self, n: int = 0, reduce_uv: bool = True, verb: bool = False) -> list_out:
        """ Verify resistance of section 'n'
            returns a list of utilization ratios
        """

        r = self.section_resistance(n=n, reduce_uv=reduce_uv)
        self.utilization[n] = {'un': r[0], 'uv': r[1], 'um': r[2], 'ut': r[3], 'ubt': r[4], 'ubc': r[5]}
        return r

    def section_resistance(self, n: int = 0, return_list: bool = True, reduce_uv: bool = True) -> (list_out, float):
        """ Calculates resistance of cross-section
            Checks the following:
                Axial force
                Shear force
                Bending moment
                Torsion
                Interaction of tension and bending moment
                Interaction of compression and bending moment
            output: r .. maximum value of different utilization ratios
        """

        UN = self.check_normal_force(n)
        UV = self.check_shear_force(n, reduce=reduce_uv)
        UM = self.check_bending_moment(n)
        UT = self.check_torsion(n)
        UBT = self.check_bending_tension(n)
        UBC = self.check_bending_compression(n)

        if return_list:
            return [UN, UV, UM, UT, UBT, UBC]
        else:
            return max([UN, UV, UM, UT, UBT, UBC])

    def check_cross_section(self):
        """ Checks if cross-section is strong enough.
            Gives boolean value for self.is_strong_enough
            Saves list of stress ratios to self.r
        """

        self.r = []
        # Cross-sectional stress ratios in member's nodes' locations
        self.r.extend(self.check_sections())
        # Buckling about y - and z - axis
        buckling_r = self.check_buckling()
        self.r.append(buckling_r[0])
        self.r.append(buckling_r[1])
        # Lateral-Torsional buckling
        self.r.append(self.check_LT_buckling())

        if max(self.r) > 1.0:
            self.is_strong_enough = False
        else:
            self.is_strong_enough = True

    def check_sections(self, reduce_uv: bool = True) -> np.ndarray:
        """ Verify resistance of all sections """
        r = np.zeros_like(self.check_section())
        for n in range(self.nsect()):
            r = np.vstack((r, self.check_section(n, reduce_uv=reduce_uv)))
            # r.append(self.check_section(n))
        return np.max(r, axis=0)

    def nsect(self) -> int:
        """ Number of sections """
        return len(self.ned)

    def add_neighbours_support(self):
        for node in self.nodes.values():
            if node is self.parent.n1 or node is self.parent.n2:
                continue
            other_members = list(filter(lambda v: v.member != self if True else False, node.parents.copy()))
            if len(other_members) == 0:
                continue

            dist_to_start = np.linalg.norm(self.parent.coordinates[0] - node.coord)
            local_coord = dist_to_start / self.length

            self.lateral_support_y.append(local_coord)

        self.calc_l_c()

    def calc_l_c(self):
        self.lateral_support_y.sort()
        self.lateral_support_z.sort()

        Lcy = []
        Lcz = []
        for i in range(len(self.lateral_support_y) + 1):
            if len(self.lateral_support_y) == 0:
                Lcy.append(self.length)
            elif i == 0:
                Lcy.append(self.length * self.lateral_support_y[i])
            elif i == len(self.lateral_support_y):
                Lcy.append(self.length * (1 - self.lateral_support_y[i - 1]))
            else:
                Lcy.append(self.length * (self.lateral_support_y[i] - self.lateral_support_y[i - 1]))
        for i in range(len(self.lateral_support_z) + 1):
            if len(self.lateral_support_z) == 0:
                Lcz.append(self.length)
            elif i == 0:
                Lcz.append(self.length * self.lateral_support_z[i])
            elif i == len(self.lateral_support_z):
                Lcz.append(self.length * (1 - self.lateral_support_z[i - 1]))
            else:
                Lcz.append(self.length * (self.lateral_support_z[i] - self.lateral_support_z[i - 1]))
        self.L_c = np.asarray([np.asarray(Lcy),
                               np.asarray(Lcz)], dtype=object)

    def design(self) -> float:
        r = max(self.check_sections())
        r = max(r, np.max(self.check_buckling()))
        if self.type in ('beam', 'rigid-beam'):
            r = max(r, self.check_LT_buckling())
            r = max(r, self.check_perpendicular_compression())

        return r

    def __repr__(self):
        return self.parent.__repr__()



# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
""" Genetic algorithm 
    Solver for optimization problems
"""


from metku.optimization.solvers import OptSolver
from metku.optimization.structopt import Variable, OptimizationProblem, \
    IndexVariable, DiscreteVariable, IntegerVariable

import random

import matplotlib.pyplot as plt
import numpy as np
from deap import base, creator, tools

import time

class GA(OptSolver):

    def __init__(self,
                 pop_size: int = 10,
                 mutation_rate: float = 0.05,
                 crossover_rate: float = 0.8,
                 penalty: callable = None,
                 mutation_fun: callable = None,
                 mutation_kwargs: dict = None,
                 crossover_fun: callable = None,
                 crossover_kwargs: dict = None,
                 select: callable = None,
                 select_kwargs: dict = None,
                 first_improvement: bool = False,
                 best_f: float = 1e100) -> None:
        """ Constructor

            All deap functions and  parameters can be found here:
            https://deap.readthedocs.io/en/master/api/tools.html

            :param pop_size: population size

            :param mutation_rate: mutation rate
            :param crossover_rate: crossover rate
            :param penalty: penalty function, arguments: constraint values
            :param mutation_fun: mutation function,
             arguments: individual chromosome
            :param mutation_kwargs: mutation function parameters
            :param crossover_fun: crossover function,
             arguments: parent chromosome1, parent chromosome2
            :param crossover_kwargs: crossover function parameters
            :param select: selection function,
             arguments: population
            :param select_kwargs: selection function parameters
            :param first_improvement:
            :param best_f: current best value
        """
        super().__init__()

        # Defaults
        if select_kwargs is None and select is None:
            select = tools.selBest
            select_kwargs = {"k": 2}
        if mutation_kwargs is None and mutation_fun is None:
            mutation_fun = tools.mutShuffleIndexes
            mutation_kwargs = {"indpb": 0.05}
        if crossover_kwargs is None and crossover_fun is None:
            crossover_fun = tools.cxTwoPoint
            crossover_kwargs = {}
        if penalty is None:
            def penalty(constr_vals):
                pos_vals = np.clip(constr_vals, 0, 100)
                return 2e6 * sum(pos_vals ** 2)

        """ Toolbox contains basic evolutionary operators """
        self.toolbox = base.Toolbox()
        self.pop_size = pop_size
        self.mutation_rate = mutation_rate
        self.crossover_rate = crossover_rate
        self.mutation_fun = mutation_fun
        self.mutation_kwargs = mutation_kwargs
        self.crossover_kwargs = crossover_kwargs
        self.crossover_fun = crossover_fun
        self.select = select
        self.select_kwargs = select_kwargs
        self.plot = False
        self.best_f = best_f
        self.prev_best = best_f
        self.first_improvement = first_improvement
        self.best_ind = None
        self.penalty = penalty
        self.counter = 0

    def take_action(self):
        pass

    def step(self, action):
        pass

    def eval_individual(self, ind):
        """
        Evaluates one individual

        :param ind: 'individual', i.e. design variable values
        :return: objective value of optimization problem
        """
        self.problem.substitute_variables(ind)
        X = [round(var.value, 3) for var in self.problem.vars]

        """ Evaluate objective function """
        obj_val = self.problem.obj(X)
        """ Evaluate constraints """
        constr_vals = self.problem.eval_cons(X)
        """ Apply penalty on constraints """
        penalty_val = self.penalty(constr_vals)

        """ Fitness function equals objective function value + penalty """
        val = obj_val + penalty_val
        if self.plot:
            self.update_plot(self.fig, self.ax)

        # print("GAPS: ", [j.g1 for j in self.problem.structure.joints.values()])

        """ If a new best solution has been found, store it """
        if val < self.best_f:
            for i, x in enumerate(X):
                ind[i] = x
            self.best_ind = ind
            self.prev_best = self.best_f
            best_x = [var.value for var in self.problem.all_vars]
            self.best_x = best_x.copy()
            self.best_f = val
            self.counter = 0

        # NOTE! Need to return at least two values, hence the comma
        return val,  # <- DO NOT REMOVE THIS COMMA

    def create_fitness(self):

        """ Create Fitness class called FitnessMin
            The weight -1.0 corresponds to minimization problem
        """
        creator.create("FitnessMin", base.Fitness, weights=(-1.0,))
        """ Create class Individual that inherits the list class and
            has the FitnessMin as its fitness attribute.
        """
        creator.create("Individual", list, fitness=creator.FitnessMin)

    def attribute_generator(self, counter_list=[0]):
        """
        Generates one gene attribute
        :param counter_list: list used for counting how
                             many times function has been called
        :return:
        """
        idx = counter_list[0] % len(self.problem.vars)
        counter_list[0] = counter_list[0] + 1

        var = self.problem.vars[idx]

        """ Creates random integer value for index or integer variables """
        if self.x0 is not None:
            val = self.x0[idx]
        elif isinstance(var, (IndexVariable, IntegerVariable)):
            val = np.random.randint(var.lb, var.ub)

        else:
            val = np.random.uniform(var.lb, var.ub)

        return val

    def register(self):
        """ Registers a set of functions to be used in the algorithm
            the arguments for the register method are:
                1. alias for the function to be registered
                2. the actual method to registered
                3. arguments to be passed to the aliased method

        """
        self.toolbox.register("attribute", self.attribute_generator)

        self.toolbox.register("individual", tools.initRepeat,
                              creator.Individual,
                              self.toolbox.attribute, len(self.problem.vars))

        # Evaluation function
        self.toolbox.register("evaluate", self.eval_individual)
        # Crossover operator
        self.toolbox.register("mate", self.crossover_fun,
                              **self.crossover_kwargs)
        # Mutation
        self.toolbox.register("mutate", self.mutation_fun,
                              **self.mutation_kwargs)

        # operator for selecting individuals for breeding the next
        # generation: each individual of the current generation
        # is replaced by the 'fittest' (best) of three individuals
        # drawn randomly from the current generation.
        # self.toolbox.register("select", tools.selBest, k=2)
        self.toolbox.register("select", self.select, **self.select_kwargs)

        # Define the population to be a list of individuals
        self.toolbox.register("population", tools.initRepeat,
                              list, self.toolbox.individual)

    def new_population(self, offspring):
        """
        Creates a new population from offsprings by shuffling them
        :param offspring: list of offsprings
        :return: new population
        """
        idx = np.random.randint(len(offspring) - 1)
        self.problem.substitute_variables(self.best_x)
        values = [var.value for var in self.problem.vars]
        for i, val in enumerate(offspring[idx]):
            offspring[idx][i] = values[i]
        multiplier = self.pop_size // len(offspring) + 1
        new_population = list(offspring) * multiplier
        np.random.shuffle(new_population[:self.pop_size])

        if self.best_ind is not None:
            new_population.append(self.best_ind)

        return new_population

    def solve(self, problem, x0=None, maxiter=100, maxtime=-1, log=False,
              min_diff=1e-5, verb=False, plot=False):


        self.problem = problem
        self.x0 = x0
        """ Creates the fitness function 
            but the create_fitness method does not return anything!
        """
        self.create_fitness()
        self.register()

        if plot and not self.plot:
            fig = plt.figure()
            ax = fig.add_subplot(111)
            plt.ion()
            plt.show()
            self.plot = True
            self.fig = fig
            self.ax = ax

        """ Create population """
        pop = self.toolbox.population(n=self.pop_size)
        if x0 is not None:
            for ind in pop:
                self.toolbox.mutate(ind)
        """ Evaluate fitness """
        fitnesses = list(map(self.toolbox.evaluate, pop))

        # CXPB  is the probability with which two individuals
        #       are crossed
        #
        # MUTPB is the probability for mutating an individual
        CXPB, MUTPB = self.crossover_rate, self.mutation_rate

        # Main loop
        it = 0
        prev_val = None
        while it < maxiter:
            it += 1
            problem.num_iters += 1
            print(f"Iteration {it}/{maxiter}")
            # Select the next generation individuals
            offspring = self.toolbox.select(pop)
            new_pop = self.new_population(offspring)

            # Clone the selected individuals
            offspring = list(map(self.toolbox.clone, new_pop))

            # Apply crossover and mutation on the offspring
            for child1, child2 in zip(offspring[::2], offspring[1::2]):
                # cross two individuals with probability CXPB
                if random.random() < CXPB:
                    self.toolbox.mate(child1, child2)
                    # fitness values of the children
                    # must be recalculated later
                    del child1.fitness.values
                    del child2.fitness.values
                for mutant in offspring[:-1]:
                    # mutate an individual with probability MUTPB
                    if random.random() < MUTPB:
                        self.toolbox.mutate(mutant)
                        del mutant.fitness.values

            # Evaluate the individuals with an invalid fitness
            invalid_ind = [ind for ind in offspring if
                           not ind.fitness.valid]
            fitnesses = map(self.toolbox.evaluate, invalid_ind)
            for ind, fit in zip(invalid_ind, fitnesses):
                ind.fitness.values = fit
            # The population is entirely replaced by the offspring
            pop[:] = offspring

            # Gather all the fitnesses in one list and print the stats
            fits = [ind.fitness.values[0] for ind in pop]
            length = len(pop)
            mean = sum(fits) / length
            sum2 = sum(x * x for x in fits)
            std = abs(sum2 / length - mean ** 2) ** 0.5

            if verb:
                print("VERB: ", self.best_x)
                print(f"Min fitness: {min(fits)} \n"
                      f"Obj: {problem.obj(self.best_x)} \n"
                      f"Max con: {max(problem.eval_cons(self.best_x)):.3f}  "
                      f"{problem.cons[np.argmax(problem.eval_cons(self.best_x))].name}")

            # Set new best values to problem
            problem.substitute_variables(self.best_x)

            if plot:
                self.update_plot(self.fig, self.ax)

            if self.first_improvement and self.best_f < self.prev_best:
                return self.best_f, self.best_x

            if self.counter > (10 - 3):
                print(
                    f"The result has not improved in the last 10 generations."
                    f"\n Returning current best values.")
                return self.best_f, self.best_x

            elif it > 0 and prev_val == min(fits):
                self.counter += 1

            else:
                prev_val = min(fits)

            if log:
                problem.num_iters += 1
                problem.fvals.append(self.best_f)
                problem.states.append(list(pop[np.argmin(fits)]))
                problem.gvals.append(max(self.constr_vals))
                self.fvals.append(self.best_f)
                self.xvals.append(self.best_x)

        return self.best_f, self.best_x


if __name__ == "__main__":
    from metku.optimization.benchmarks import *

    problem = TenBarTruss('index')
    solver = GA(pop_size=100)
    solver.solve(problem, maxiter=50)
